package Newslash::Command::configimport;
use Mojo::Base 'Newslash::Command';

use Mojo::Util 'getopt';
use Data::Dumper;
use Encode;
use JSON qw(decode_json);
use IO::File;
use List::Util qw(any);

has description => 'Import some configs form files';
has usage => sub { shift->extract_usage };

has experimental => 1;
has danger => 1;

has experimental => 1;
has danger => 1;

use constant TYPES => qw(message_type);

sub run {
    my ($self, @args) = @_;
    return if !$self->check_force_option(\@args);

    getopt \@args,
      't|type=s' => \my $type,
      'j|jsonfile=s' => \my $filename;

    if (!$type || ! any { $_ eq $type } (TYPES)) {
        $self->app->log->error("configimport: no config type given!");
        return;
    }
    if (!$filename) {
        $self->app->log->error("configimport: no config file given!");
        return;
    }
    if (! -r $filename) {
        $self->app->log->error("configimport: $filename not exist or not readable!");
        return;
    }

    if ($type eq "message_type") {
        $self->{db} = $self->app->db;
        $self->_create_table();
        if (!$self->_insert_message_type($filename)) {
            $self->app->log->error("configimport: import $type from $filename failed!");
            return;
        }
    }
    else {
        $self->app->log->error("configimport: $type is unimplemented.");
        return;
    }
    $self->app->log->info("configimport: import $type from $filename done.");
}

sub _create_table {
    my ($self) = @_;
    my $table_name = "ns_message_types";
    my $dbh = $self->{db}->connect_db;

    my $sql = <<"EOSQL";
CREATE TABLE IF NOT EXISTS $table_name (
  message_type_id     int(11)     NOT NULL PRIMARY KEY,
  name                varchar(32) NOT NULL,
  description         varchar(32) NOT NULL,
  seclev              int(11)     NOT NULL DEFAULT '1',
  selectable_mode     int(11)     NOT NULL DEFAULT '1'
) DEFAULT CHARSET=utf8mb4
EOSQL
    my $rs = $dbh->do($sql, undef);
    $self->{db}->disconnect_db;

    if ($rs) {
        return 1;
    }
    $self->app->log->info("configimport: table '$table_name' not created.");
    return;
}

sub _insert_message_type {
    my ($self, $json_file) = @_;
    my $table_name = "ns_message_types";
    my $dbh = $self->{db}->start_transaction;

    my $fh = IO::File->new($json_file, "r");
    return if !$fh;

    my $json_text = do { local $/; <$fh> };
    $fh->close;
    my $types = decode_json($json_text);

    my $sql = <<"EOSQL";
INSERT INTO $table_name 
    (message_type_id, name, description, seclev, selectable_mode)
  VALUES
    (?,               ?,    ?,           ?,      ?)
EOSQL
    my $sth = $dbh->prepare($sql);
    for my $item (@{$types->{items}}) {
        my $rs = $sth->execute($item->{id},
                               $item->{name},
                               $item->{description},
                               $item->{seclev},
                               $item->{selectable});
        if (!$rs) {
            $self->app->log->error("configimport: database error: " . $dbh->errstr);
            $self->{db}->rollback;
            return;
        }
    }
    $self->{db}->commit;
    return 1;
}

1;

=encoding utf8

=head1 NAME

Newslash::Command::configimport - newslash administer command (configimport)

=head1 SYNOPSIS

  Usage: APPLICATION configimport [OPTIONS]

      ./newslash_web databaseinit -t message_type -j messagae_type.json

    Options:
      -f  if mode is not "test", need this option
      -t  configuration type
      -j  configuration JSON file

=head1 DESCRIPTION

  L<Newslash::Command::configimport> imports configuration from JSON files

=head1 ATTRIBUTES

L<Newslash::Command::configimport> inherits all attributes from
L<Newslash::Command> and implements the following new ones.

=head2 description

  my $description = $daemon->description;
  $daemon         = $daemon->description('Foo');

Short description of this command, used for the command list.

=head2 usage

  my $usage = $daemon->usage;
  $daemon   = $daemon->usage('Foo');

Usage information for this command, used for the help screen.

=head1 METHODS

L<Newslash::Command::configimport> inherits all methods from
L<Newslash::Command> and implements the following new ones.

=head2 run

  $daemon->run(@ARGV);

Run this command.

=head1 SEE ALSO

L<Mojolicious>, L<Mojolicious::Guides>, L<https://mojolicious.org>.

=cut
