package Newslash::Model::Boxes;
use Newslash::Model::Base -base;

use constant TABLE_NAME => 'ns_boxes';

sub create_table {
    my $self = shift;
    my $dbh = $self->connect_db;
    my $table_name = TABLE_NAME;

    if ($self->check_readonly) {
        return 1;
    }

    my $sql = <<"EOSQL";
CREATE TABLE IF NOT EXISTS $table_name (
  id             int(8) unsigned AUTO_INCREMENT PRIMARY KEY ,
  name           varchar(128)    NOT NULL UNIQUE,
  security_level smallint        NOT NULL DEFAULT 0,
  title          varchar(256)    NOT NULL DEFAULT "",
  comment        varchar(256)    NOT NULL DEFAULT "",
  type           varchar(64)     NOT NULL DEFAULT "",
  model          varchar(128)    NOT NULL,
  query_params   mediumtext      NOT NULL,
  `limit`        int             DEFAULT 5,
  template       mediumtext      NOT NULL
) DEFAULT CHARSET=utf8mb4
EOSQL
    my $rs = $dbh->do($sql, undef);
    $self->disconnect_db;
    if ($rs) {
        return 1;
    }
    return;
}

sub on_start_up {
    my $self = shift;
    if (!$self->table_exists(TABLE_NAME)) {
        $self->logger->warn("Boxes: table not exists, create it.");
        return $self->create_table;
    }
    return 1;
}

sub select {
    my $self = shift;
    my $params = {@_};
    my $table_name = TABLE_NAME;

    my $target = '';
    $target = 'name' if ($params->{name});
    $target = 'id' if ($params->{id});

    if ($target) {
        my $dbh = $self->connect_db;
        my $sql = <<"EOSQL";
SELECT * FROM $table_name
  WHERE $target = ?
EOSQL
        my $sth = $dbh->prepare($sql);
        $sth->execute($params->{$target});
        my $rs = $sth->fetchall_arrayref({});
        return if @$rs == 0;
        return $rs->[0];
    }

    # return all item
    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
SELECT * FROM $table_name
  ORDER BY id ASC
EOSQL
    my $sth = $dbh->prepare($sql);
    $sth->execute;
    my $rs = $sth->fetchall_arrayref({});
    if (!$rs) {
        $self->set_error($dbh->err, $dbh->errno);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;
    return $rs;
}

sub create {
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $table_name = TABLE_NAME;

    my $err = '';
    if (!$params->{name}) {
        $err = "no item name";
    }
    if (length $err) {
        $self->set_error($err, -1);
        return;
    }

    my $sql = <<"EOSQL";
INSERT INTO $table_name
    (name, security_level, title, comment, type, model, query_params, `limit`, template)
  VALUES
    (?,    ?,              ?,     ?,       ?,    ?,     ?,            ?,       ?)
EOSQL
    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef,
                      $params->{name},
                      $params->{security_level} || 0,
                      $params->{title} || "",
                      $params->{comment} || "",
                      $params->{type} || "",
                      $params->{model} || "",
                      $params->{query_params} || "",
                      $params->{limit} || 5,
                      $params->{template} || "");
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    my $id = $dbh->last_insert_id(undef, undef, undef, undef);
    $self->disconnect_db;

    return $id;
}

sub update {
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $table_name = TABLE_NAME;

    my $message = '';
    $message = "no_id_given" if !$params->{id};

    if (length $message) {
        $self->set_error($message, -1);
        return;
    }

    my @cols;
    my @attrs;
    for my $k (qw|name security_level title comment type model query_params limit template|) {
        if (defined $params->{$k}) {
            push @cols, "`$k` = ?";
            push @attrs, $params->{$k};
        }
    }
    my $columns = join(", ", @cols);

    my $sql = <<"EOSQL";
UPDATE $table_name
  SET $columns
  WHERE id = ?
EOSQL

    push @attrs, $params->{id};

    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef, @attrs);
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;

    return $params->{id};
}

sub delete {
    my $self = shift;
    return if $self->check_readonly;

    my $params = {@_};
    my $table_name = TABLE_NAME;

    if (!$params->{id}) {
        $self->set_error("no id given");
        return;
    }

    my $sql = <<"EOSQL";
DELETE FROM $table_name
  WHERE id = ?
EOSQL
    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef, $params->{id});
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->disconnect_db;
        return;
    }
    $self->disconnect_db;
    return 1;
}

sub export_all {
    my $self = shift;
    my $target = shift;
    my $table_name = TABLE_NAME;

    if (ref($target) eq "SCALAR") {
        $$target = $self->export_json(table => $table_name,
                                      sort_key => "name",
                                      exclude => [qw(id)],
                                     );
        return 1;
    }
    my $output = $self->export_json(file => $target,
                                    table => $table_name,
                                    sort_key => "name",
                                    exclude => [qw(id)],
                                   );
    return if !$output;
    return 1;
}

sub import_all {
    my $self = shift;
    my $target = shift;
    my $table_name = TABLE_NAME;

    my $output;

    if (ref($target) eq "SCALAR") {
        my $rs = $self->import_json(table => $table_name,
                                    json => $$target,
                                    unique_key => "name",
                                    exclude => [qw(id)]
                                     );
        return $rs;
    }
    my $rs = $self->import_json(table => $table_name,
                                file => $target,
                                unique_key => "name",
                                exclude => [qw(id)],
                               );
    return $rs;
}



1;
