package Newslash::Model::Firehose;
use Newslash::Model::Base -base;
use Data::Dumper;
use List::Util qw(any);

sub key_definition {
    return {
            table => "firehose",
            primary => "id",
            unique => ["globjid"],
            datetime => [qw(createtime last_update)],
            timestamp => "createtime",
            addable => [qw(popularity editorpop neediness
                         activity)],
            other => [qw(uid discussion type
                         accepted rejected
                         public attention_needed
                         is_spam bayes_spam collateral_spam
                         primaryskid tid
                         srcid url_id toptags
                         email emaildomain name dept
                         ipid subnetid category nexuslist
                         signoffs stoid
                         body_length word_count
                         srcname thumb mediatype
                         offmainpage sprite sprite_info
                         preview tags.tagnameid globj_types.maintable
                       )],
            aliases => { firehose_id => "id",
                         globj_id => "globjid",
                         src_id => "srcid",
                         create_time => "createtime",
                         update_time => "updatetime",
                         tag_name_id => "tags.tagnameid",
                         tagname_id => "tags.tagnameid",
                         tagnameid => "tags.tagnameid",
                         maintable => "globj_types.maintable",
                       }
           };
}

sub select {
    my $self = shift;
    my $params = {@_};

    # get tag_name_id from tagname
    if ($params->{tagname} || $params->{tag_name}) {
        my $tag_name = $params->{tagname} || $params->{tag_name};
        my $tags = $self->new_instance_of("Tags");
        my $rs = $tags->select_tagnames(tagname => $tag_name);
        if ($rs) {
            $params->{tag_name_id} = $rs->{tagnameid};
        }
        else {
            return [];
        }
    }

    my ($where_clause, $where_values, $unique) = $self->build_where_clause(params => $params);
    my ($limit_clause, $limit_values) = $self->build_limit_clause(params => $params, default => 50);
    my ($orderby_clause, $orderby_values) = $self->build_order_by_clause(params => $params);

    # show future story?
    my @where_clauses;
    if ($params->{hide_future}) {
        push @where_clauses, "firehose.createtime <= NOW()";
    }

    # show non-public story?
    if ($params->{public_only}) {
        push @where_clauses, "firehose.public != 'no'";
    }

    if (@where_clauses) {
        if ($where_clause) {
            $where_clause = $where_clause . " AND ";
        }
        else {
            $where_clause = "WHERE ";
        }
        $where_clause = $where_clause . join(" AND ", @where_clauses);
    }

    my @attrs;
    push @attrs, @$where_values, @$limit_values, @$orderby_values;


    my $join_clause = "";
    if ($params->{tag_name_id} || $params->{tagnameid}) {
        $join_clause = "LEFT JOIN tags ON firehose.globjid = tags.globjid";
    }

    my $dbh = $self->connect_db;
    my $sql = <<"EOSQL";
SELECT DISTINCT firehose.id AS firehose_id, firehose.*, firehose_text.*, topics.tid, stories.sid,
       users.nickname AS author, urls.url, globj_types.maintable
  FROM firehose
    LEFT JOIN topics ON firehose.tid = topics.tid
    LEFT JOIN firehose_text ON firehose.id = firehose_text.id
    LEFT JOIN stories ON firehose.type = "story" AND firehose.srcid = stories.stoid
    LEFT JOIN urls ON firehose.type = "submission" AND firehose.url_id = urls.url_id
    LEFT JOIN users ON firehose.uid = users.uid
    LEFT JOIN globjs ON firehose.globjid = globjs.globjid
    LEFT JOIN globj_types ON globjs.gtid = globj_types.gtid
    $join_clause
    $where_clause
    $orderby_clause
    $limit_clause
EOSQL

    #warn $sql;
    #warn Dumper @attrs;

    my $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $items = $sth->fetchall_arrayref({});

    if (!$items) {
        $self->disconnect_db;
        return;
    }
    if (@$items == 0) {
        $self->disconnect_db;
        return $unique ? undef : [];
    }

    # get tags
    $sql = <<"EOSQL";
SELECT tags.*, tagnames.tagname
  FROM ( SELECT firehose.globjid, firehose_text.*, topics.tid
           FROM firehose
             LEFT JOIN topics ON firehose.tid = topics.tid
             LEFT JOIN firehose_text ON firehose.id = firehose_text.id
             LEFT JOIN globjs ON firehose.globjid = globjs.globjid
             LEFT JOIN globj_types ON globjs.gtid = globj_types.gtid
             $join_clause
             $where_clause
             $orderby_clause
             $limit_clause
        ) AS target
    LEFT JOIN tags ON target.globjid = tags.globjid
    LEFT JOIN tagnames ON tags.tagnameid = tagnames.tagnameid
EOSQL

    $sth = $dbh->prepare($sql);
    $sth->execute(@attrs);
    my $tags_table = $sth->fetchall_arrayref({});

    my $tags = {};
    for my $tag (@$tags_table) {
        my $globjid = $tag->{globjid};
        if (!$tags->{$globjid}) {
            $tags->{$globjid} = [];
        }
        push @{$tags->{$globjid}}, $tag;
    }

    for my $item (@$items) {
        my $globjid = $item->{globjid};
        $item->{tags} = $tags->{$globjid} if $tags->{$globjid};
        $self->_generalize($item);
    }

    return $items->[0] if $unique;
    return $items;
}

sub update {
    my $self = shift;
    my $params = {@_};

    # keys for firehose_text table
    my $text_params = {};
    $text_params->{id} = $params->{id} || $params->{firehose_id};
    $text_params->{title} = $params->{title};
    $text_params->{introtext} = $params->{intro_text} || $params->{introtext};
    $text_params->{bodytext} = $params->{body_text} || $params->{bodytext};
    $text_params->{media} = $params->{media};

    if (!defined $text_params->{title} &&
        !defined $text_params->{introtext} &&
        !defined $text_params->{bodytext} &&
        !defined $text_params->{media}) {
        return $self->generic_update(params => $params);
    }

    $self->start_transaction;

    # update firehose_text table
    my $rs = $self->generic_update(table => "firehose_text",
                                   updatable_keys => [qw(title introtext bodytext media)],
                                   condition_keys => { "id" => "firehose_text.id" },
                                   params => $text_params);
    if (!defined $rs) {
        $self->rollback;
        return;
    }

    if ($rs > 0) {
        $params->{last_update} => { function => "NOW()" };
    }

    my $rs2 = $self->generic_update(params => $params);
    if (!defined $rs2) {
        $self->rollback;
        return;
    }

    $self->commit;
    return $rs > 0 ? $rs : $rs2;

}


sub _generalize {
    my ($self, $item) = @_;

    # NTO-nized
    $item->{globj_id} = $item->{globjid};
    $item->{id} = $item->{firehose_id};
    $item->{create_time} = $item->{createtime};
    $item->{update_time} = $item->{last_update};

    $item->{primary_topic} = {};
    $item->{primary_topic}->{tid} = $item->{tid};
    for my $k (qw{keyword textname series image width height
                  submittable searchable storypickable usesprite}) {
        next if !$item->{$k};
        $item->{primary_topic}->{$k} = $item->{$k};
        delete $item->{$k};
    }

    $item->{content_type} = $item->{type};
    $item->{intro_text} = $item->{introtext};
    $item->{bodytext} ||= "";
    $item->{body_text} = $item->{bodytext};

    if ($item->{body_text}) {
        $item->{full_text} = join("\n", $item->{intro_text}, $item->{body_text});
    }
    else {
        $item->{full_text} = $item->{intro_text};
    }
    $item->{fulltext} = $item->{full_text};

    $item->{discussion_id} = $item->{discussion};

    # no public flag given, public is 'yes'
    $item->{public} = 'yes' if !$item->{public};
}


## WORNING:
## Below is legacy from slash. deprecated, and you can use only backword-compatible.

# getEntryPopularityForColorLevel - import from FireHose.pm
sub getEntryPopularityForColorLevel {
    my ($self, $level) = @_;

    # my $midpop = $self->getEntryPopularityForColorLevel(5);
    # argument is 5, then value is "102, 93". entry is 102, min is 93
    # '290,240 220,200 185,175 155,138 102,93 30,25 0,-20 -60,-999999'
    my $slicepoints = '290,240 220,200 185,175 155,138 102,93 30,25 0,-20 -60,-999999';
    my @levels = split / /, $slicepoints;
    my $entry_min = $levels[$level-1];
    my ($entry, $min) = split /,/, $entry_min;
    return $entry;
}

sub createFireHose {
    my ($self, $data) = @_;
    $data->{dept} ||= "";
    $data->{discussion} = 0 if !defined $data->{discussion} || !$data->{discussion};
    $data->{-createtime} = "NOW()" if !$data->{createtime} && !$data->{-createtime};
    $data->{discussion} ||= 0 if defined $data->{discussion};
    $data->{popularity} ||= 0;
    $data->{editorpop} ||= 0;
    $data->{body_length} ||= $data->{bodytext} ? length($data->{bodytext}) : 0;
    $data->{word_count} = countWords($data->{introtext}) + countWords($data->{bodytext});
    $data->{mediatype} ||= "none";
    $data->{email} ||= '';

    my $text_data = {};
    $text_data->{title} = delete $data->{title};
    $text_data->{introtext} = delete $data->{introtext};
    $text_data->{bodytext} = delete $data->{bodytext};
    $text_data->{media} = delete $data->{media};

    my $db = $self->new_instance_of('Newslash::Model::LegacyDB');
    #$self->sqlDo('SET AUTOCOMMIT=0');
    $db->sqlDo('SET AUTOCOMMIT=0');

    #my $ok = $self->sqlInsert("firehose", $data);
    my $ok = $db->sqlInsert("firehose", $data);
    if (!$ok) {
        warn "could not create firehose row, '$ok'";
    }
    if ($ok) {
        #$text_data->{id} = $self->getLastInsertId({table => 'firehose', prime => 'id'});
        $text_data->{id} = $db->getLastInsertId({table => 'firehose', prime => 'id'});

        #$ok = $self->sqlInsert("firehose_text", $text_data);
        $ok = $db->sqlInsert("firehose_text", $text_data);
        if (!$ok) {
            warn "could not create firehose_text row for id '$text_data->{id}'";
        }
    }

    if ($ok) {
        #$self->sqlDo('COMMIT');
        $db->sqlDo('COMMIT');
    } else {
        #$self->sqlDo('ROLLBACK');
        $db->sqlDo('ROLLBACK');
    }
    #$self->sqlDo('SET AUTOCOMMIT=1');
    $db->sqlDo('SET AUTOCOMMIT=1');

    # set topics rendered appropriately
    # if ($ok) {
    #     if ($data->{type} eq "story") {
    #         #my $tids = $self->sqlSelectColArrayref("tid", "story_topics_rendered", "stoid='$data->{srcid}'");
    #         my $tids = $db->sqlSelectColArrayref("tid", "story_topics_rendered", "stoid='$data->{srcid}'");
    #         $self->setTopicsRenderedForStory($data->{srcid}, $tids);
    #     } else {
    #         $self->setTopicsRenderedBySkidForItem($text_data->{id}, $data->{primaryskid});
    #     }
    # }

    return $text_data->{id};
}

sub setFireHose {
    my($self, $id, $data) = @_;
    #my $constants = getCurrentStatic();
    return undef unless $id && $data;
    return 0 if !%$data;
    my $db = $self->new_instance_of('LegacyDB');
    #my $id_q = $self->sqlQuote($id);
    my $id_q = $db->sqlQuote($id);

    #my $mcd = $self->getMCD();
    #my $mcdkey;
    #if ($mcd) {
    #    $mcdkey = "$self->{_mcd_keyprefix}:firehose";
    #}

    if (!exists($data->{last_update}) && !exists($data->{-last_update})) {
        my @non_trivial = grep {!/^activity$/} keys %$data;
        if (@non_trivial > 0) {
            $data->{-last_update} = 'NOW()';
        } else {
            $data->{-last_update} = 'last_update';
        }
    }

    # Admin notes used to be stored in firehose.note;  that column is
    # now gone and that data goes in globj_adminnote.  The note is
    # stored on the object that the firehose points to.
    if (exists $data->{note}) {
        my $note = delete $data->{note};
        # XXX once getFireHose does caching, use that instead of an sqlSelect
        #my $globjid = $self->sqlSelect('globjid', 'firehose', "id=$id_q");
        my $globjid = $db->sqlSelect('globjid', 'firehose', "id=$id_q");
        warn "no globjid for firehose '$id'" if !$globjid;
        my $globjs = $self->new_instance_of('Globjs');
        #$self->setGlobjAdminnote($globjid, $note);
        $globjs->setGlobjAdminnote($globjid, $note);
    }

    return 0 if !keys %$data;

    my $text_data = {};

    $text_data->{title} = delete $data->{title} if exists $data->{title};
    $text_data->{introtext} = delete $data->{introtext} if exists $data->{introtext};
    $text_data->{bodytext} = delete $data->{bodytext} if exists $data->{bodytext};
    $text_data->{media} = delete $data->{media} if exists $data->{media};

    #my $rows = $self->sqlUpdate('firehose', $data, "id=$id_q");
    my $rows = $db->sqlUpdate('firehose', $data, "id=$id_q");
    #{ use Data::Dumper; my $dstr = Dumper($data); $dstr =~ s/\s+/ /g; print STDERR "setFireHose A rows=$rows for id=$id_q data: $dstr\n"; }
    #$rows += $self->sqlUpdate('firehose_text', $text_data, "id=$id_q") if keys %$text_data;
    $rows += $db->sqlUpdate('firehose_text', $text_data, "id=$id_q") if keys %$text_data;
    #{ use Data::Dumper; my $dstr = Dumper($text_data); $dstr =~ s/\s+/ /g; print STDERR "setFireHose B rows=$rows for id=$id_q data: $dstr\n"; }

    if (defined $data->{primaryskid}) {
        my $type = $data->{type};
        if (!$type) {
            my $item = $self->getFireHose($id);
            $type = $item->{type};
        }
        if ($type ne "story") {
            $self->setTopicsRenderedBySkidForItem($id, $data->{primaryskid});
        }
    }

	$self->deleteFireHoseCaches([ $id ], 1);

	return $rows;
}

# getFireHose and getFireHoseMulti are the standard ways to retrieve
# a firehose item's data, given its firehose id.  It is recommended
# that all item data retrieval bottleneck through here, to take full
# advantage of caching.

sub getFireHose {
    my($self, $id, $options) = @_;
    if ($id !~ /^\d+$/) {
        return undef;
    }
    my $db = $self->new_instance_of('LegacyDB');
    #my $hr = $self->getFireHoseMulti([$id], $options);
    my $hr = $db->getFireHoseMulti([$id], $options);
    return $hr->{$id};
}

# Like getFireHose, but does the lookup by globjid (which is unique).
#
# One way to implement this would be to store complete hose data with
# each globjid, the advantage being half as many memcached requests
# and thus half the latency.  I wouldn't mind paying most of the price
# for this:  extra memcached RAM, and extra network bandwidth in some
# cases.  But that would also interleave hose item cache expiration
# times, which could lead to confusing bugs.  To avoid that, I'm
# willing to pay the price of extra latency.
#
# So instead this does a lookup from globjid to id (which is cached),
# and wraps that conversion about a getFireHoseMulti call.
#
# The option id_only skips the step of retrieving the actual firehose
# data and returns a hashref (or values of hashrefs) with only the
# id field populated.

sub getFireHoseByGlobjid {
    my($self, $globjid, $options) = @_;
    my $hr = $self->getFireHoseByGlobjidMulti([$globjid], $options);
    return $hr->{$globjid};
}

sub getFireHoseByGlobjidMulti {
    my($self, $globjid_ar, $options) = @_;
    my $ret_hr = { };

	# First, convert the globjids to ids.

    #my $exptime = 86400;
    my $globjid_to_id_hr = { };

    my $mcd_hr = { };
    #my $mcd = $self->getMCD();
    #my $mcdkey;
    #if ($mcd && !$options->{memcached_no_read}) {
    #    $mcdkey = "$self->{_mcd_keyprefix}:gl2id" if $mcd;
    #    my @keylist = ( map { "$mcdkey:$_" } @$globjid_ar );
    #    my $mcdkey_hr = $mcd->get_multi(@keylist);
    #    for my $k (keys %$mcdkey_hr) {
    #        my($id) = $k =~ /^\Q$mcdkey:\E(\d+)$/;
    #        next unless $id;
    #        $mcd_hr->{$id} = $mcdkey_hr->{$k};
    #    }
    #}

    my $answer_hr = { };
    my @remaining_ids = ( grep { !defined($mcd_hr->{$_}) } @$globjid_ar );
    my $splice_count = 2000;
    while (@remaining_ids) {
        my @globjid_chunk = splice @remaining_ids, 0, $splice_count;
        my $globjid_str = join(',', @globjid_chunk);
        my $more_hr = $self->sqlSelectAllKeyValue(
                                                  'globjid, id',
                                                  'firehose',
                                                  "globjid IN ($globjid_str)");
        # XXX faster, or slower, than iterating over $more_hr?
        $answer_hr = {( %$answer_hr, %$more_hr )};
    }

    #if ($mcd && %$answer_hr && !$options->{memcached_no_write}) {
    #    for my $globjid (keys %$answer_hr) {
    #        $mcd->set("$mcdkey:$globjid", $answer_hr->{$globjid}, $exptime);
    #    }
    #}

    $globjid_to_id_hr = {( %$mcd_hr, %$answer_hr )};

    # If only the ids are desired, we can return those now.
    if ($options->{id_only}) {
        return $globjid_to_id_hr;
    }

	# Now that the globjids have been converted to ids, call
	# getFireHoseMulti.

    my @ids = grep { $_ } map { $globjid_to_id_hr->{$_} } @$globjid_ar;
    my $firehose_hr = $self->getFireHoseMulti(\@ids, $options);
    use Data::Dumper; print STDERR scalar(gmtime) . " $$ gFHM returned non-hashref from '@ids' options: " . Dumper($options) if ref($firehose_hr) ne 'HASH';

    # Then convert the keys in the answer back to globjids.
    for my $id (keys %$firehose_hr) {
        if (ref($firehose_hr->{$id}) ne 'HASH') {
            my $opt_str = Dumper($options); $opt_str =~ s/\s+/ /g;
            my $fh_str = Dumper($firehose_hr); $fh_str =~ s/\s+/ /g;
            print STDERR scalar(gmtime) . " $$ gFHM returned non-hashref-hashref for '$id' from '@ids' options '$opt_str' fh '$fh_str'\n";
            next;
        }
        $ret_hr->{ $firehose_hr->{$id}{globjid} } = $firehose_hr->{$id};
    }
    return $ret_hr;
}

sub getFireHoseIdFromGlobjid {
    my($self, $globjid) = @_;
    return $self->getFireHoseByGlobjid($globjid, { id_only => 1 });
}

sub getFireHoseIdFromUrl {
    my($self, $url) = @_;
    if ($url) {
        my $fudgedurl = fudgeurl($url);
        if ($fudgedurl) {
            my $url_id = $self->getUrlIfExists($fudgedurl);
            if ($url_id) {
                return $self->getPrimaryFireHoseItemByUrl($url_id);
            }
        }
    }
    return 0;
}

sub setTopicsRenderedBySkidForItem {
    my($self, $id, $primaryskid) = @_;
    #my $constants = getCurrentStatic();
    my $db = $self->new_instance_of('LegacyDB');
    my $skins = $self->new_instance_of('Skins');

    #my $skin = $self->getSkin($primaryskid);
    my $skin = $skins->getSkin($primaryskid);

    # if no primaryskid assign to mainpage skid
    #my $nexus = $skin && $skin->{nexus} ? $skin->{nexus} : $constants->{mainpage_nexus_tid};
    my $nexus = $skin && $skin->{nexus} ? $skin->{nexus} : 49;

    #$self->sqlDelete("firehose_topics_rendered", "id = $id");
    #$self->sqlInsert("firehose_topics_rendered", { id => $id, tid => $nexus });
    #$self->setFireHose($id, { nexuslist => " $nexus " });
    $db->sqlDelete("firehose_topics_rendered", "id = $id");
    $db->sqlInsert("firehose_topics_rendered", { id => $id, tid => $nexus });
    $db->setFireHose($id, { nexuslist => " $nexus " });
}

sub setTopicsRenderedForStory {
    my($self, $stoid, $tids) = @_;
    my $the_tids = [ @$tids ]; # Copy tids so any changes we make don't affect the caller
    #my $constants = getCurrentStatic();
    my $globjs = $self->new_instance_of('Globjs');
    my $slash_db = $self->new_instance_of('SlashDB');

    my $story = $slash_db->getStory($stoid, "", 1);
    if ($story) {
        my $globjid = $globjs->getGlobjidCreate("stories", $story->{stoid});
        my $id = $self->getFireHoseIdFromGlobjid($globjid);
        my @nexus_topics;
        if ($id) {
            $slash_db->sqlDelete("firehose_topics_rendered", "id = $id");
            foreach (@$the_tids) {
                $slash_db->sqlInsert("firehose_topics_rendered", { id => $id, tid => $_});
            }
            my $tree = $slash_db->getTopicTree();
            @nexus_topics = grep { $tree->{$_}->{nexus} } @$the_tids;
            my $nexus_list = join ' ', @nexus_topics;
            $nexus_list = " $nexus_list ";
            $self->setFireHose($id, { nexuslist => $nexus_list });
        }
    }
}

sub hard_delete {
    my ($self, $type, $srcid) = @_;
    my $error = 0;

    my $targets = $self->select(type => $type,
                               src_id => $srcid,
                              );
    if (!$targets) {
        Mojo::Log->new->warn("Firehose::hard_delete: target not found. type: $type, src_id: $srcid");
        return 1;
    }
    my $dbh = $self->start_transaction;

    # delete firehose related table items

    my $sql = "DELETE FROM firehose WHERE id = ?";
    my $rs;

    for my $target (@$targets) {
        $rs = $dbh->do($sql, undef, $target->{id});
        if (!defined $rs || $rs == 0) {
            Mojo::Log->new->warn("DELETE FROM firehose failed. id is $target->{id}");
            $error = 1;
        }

        $sql = "DELETE FROM firehose_text WHERE id = ?";
        $rs = $dbh->do($sql, undef,  $target->{id});
        if (!defined $rs) {
            Mojo::Log->new->warn("DELETE FROM firehose_text failed. id is $target->{id}");
            $error = 1;
        }

        $sql = "DELETE FROM firehose_topics_rendered WHERE id = ?";
        $rs = $dbh->do($sql, undef,  $target->{id});
        if (!defined $rs) {
            Mojo::Log->new->warn("DELETE FROM firehose_topics_rendered failed. id is $target->{id}");
            $error = 1;
        }

        # delete tags
        $sql = "DELETE FROM tags WHERE globjid = ?";
        $rs = $dbh->do($sql, undef, $target->{globjid});
        if (!defined $rs) {
            Mojo::Log->new->warn("DELETE FROM tags failed. globjid is $target->{globjid}");
            $error = 1;
        }

        # delete globjs
        $sql = "DELETE FROM globjs WHERE globjid = ?";
        $rs = $dbh->do($sql, undef, $target->{globjid});
        if (!defined $rs) {
            Mojo::Log->new->warn("DELETE FROM globjs failed. globjid is $target->{globjid}");
            $error = 1;
        }
    };

    $self->commit;
    return !$error;

}

1;
