package Newslash::Model::Messages;
use Newslash::Model::Base -base;

use constant MESSAGE_TYPES_TABLE_NAME => 'ns_message_types';

use JSON qw(decode_json);
use IO::File;

sub on_start_up {
    my $self = shift;
    my $app = shift;

    # create tables
    return 1 if $self->check_readonly;
    my $table_name = MESSAGE_TYPES_TABLE_NAME;
    if (!$self->table_exists($table_name)) {
        my $dbh = $self->connect_db;

        my $sql = <<"EOSQL";
CREATE TABLE IF NOT EXISTS $table_name (
  message_type_id     int(11)     NOT NULL PRIMARY KEY,
  name                varchar(32) NOT NULL,
  description         varchar(32) NOT NULL,
  seclev              int(11)     NOT NULL DEFAULT '1',
  selectable_mode     int(11)     NOT NULL DEFAULT '1'
) DEFAULT CHARSET=utf8mb4
EOSQL
        my $rs = $dbh->do($sql, undef);
        $self->disconnect_db;
        if ($rs) {
            return 1;
        }
        $self->warn("cannot create table: $table_name");
    }

    return 1;
}

sub create_message_type {
    my ($self, $id, $name, $description, $seclev, $selectable) = @_;
    return if $self->check_readonly;
    return if (!defined $id || !$name || !$description);

    my $table_name = MESSAGE_TYPES_TABLE_NAME;
    my $sql = <<"EOSQL";
INSERT INTO $table_name 
    (message_type_id, name, description, seclev, selectable_mode)
  VALUES
    (?,               ?,    ?,           ?,      ?)
EOSQL
    my $dbh = $self->connect_db;
    my $sth = $dbh->prepare($sql);
    my $rs = $sth->execute($id, $name, $description, $seclev, $selectable);
    if (defined $rs) {
        $self->disconnect_db;
        return $id;
    }
    $self->disconnect_db;
    return $rs;
};

sub delete_message_type {
    my $self = shift;
    return if $self->check_readonly;
    my $params = {@_};
    return if !defined $params->{message_type_id};

    my $table_name = MESSAGE_TYPES_TABLE_NAME;
    my $sql = <<"EOSQL";
DELETE FROM $table_name WHERE message_type_id = ?
EOSQL
    my $dbh = $self->connect_db;
    my $rs = $dbh->do($sql, undef, $params->{message_type_id});
    $self->disconnect_db;
    return $rs;
};

sub select_message_type {
    return shift->select(@_);
}

sub select {
    my $self = shift;
    my $table_name = MESSAGE_TYPES_TABLE_NAME;
    return $self->generic_select(table => $table_name,
                                 unique_keys => [qw(message_type_id name)],
                                 params => {@_});
}

sub create {
    my $self = shift;
    my $params = {@_};
    my $uid = $params->{to};
    my $from_uid = $params->{from};
    my $title = $params->{title};
    my $body = $params->{body};
    my $code = $params->{code};
    my $type = $params->{type} || "";

    my $users = $self->new_instance_of("Users");

    return if $self->check_readonly;
    return if (!$uid || !$title || !$body || !$code);

    # get message mode
    my $message_configs = $users->messages->select(uid => $uid);
    my $mode = -1;
    for my $conf (@$message_configs) {
        if ($conf->{code} == $code) {
            $mode = $conf->{mode};
            last;
        }
    }
    return 0 if $mode == -1;

    # create message_drop
    my $dbh = $self->start_transaction;
    my $sql = <<"EOSQL";
INSERT INTO message_drop
    (user, fuser, code, date,  altto, send,    message)
  VALUE
    (?,    ?,     ?,    NOW(), "",    "defer", ?)
EOSQL

    my $rs = $dbh->do($sql, undef, $uid, $from_uid, $code, $body);
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }
    my $message_id = $dbh->last_insert_id(undef, undef, undef, undef);

    # create log
    $sql = <<"EOSQL";
INSERT INTO message_log
    (id, user, fuser, code, mode, date)
  VALUE
    (?,  ?,    ?,     ?,    ?,    NOW())
EOSQL

    $rs = $dbh->do($sql, undef,
                      $message_id,
                      $uid,
                      $from_uid,
                      $code,
                      $mode);
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }

    # send message
    if ($mode == 0) {
        # mode == 0: mail
        #my $mail = $self->new_instance_of("mail_messages");
        #return $mail->create($uid, $title, $body, $code, $options);
    }
    elsif ($mode == 1) {
        # mode == 1: web
        my $sender = ($type eq "notify") ? 0 : $from_uid;
        # send
        my $web = $self->new_instance_of("WebMessages");
        my $rs = $web->create($message_id, $uid, $sender, $title, $body, $code);
        if (!$rs) {
            $self->set_error($dbh->errstr, $dbh->err);
            $self->rollback;
            return;
        }
    }

    # delete message_drop
    $sql = <<"EOSQL";
DELETE FROM message_drop WHERE id = ?
EOSQL
    $rs = $dbh->do($sql, undef,
                   $message_id);
    if (!$rs) {
        $self->set_error($dbh->errstr, $dbh->err);
        $self->rollback;
        return;
    }

    $self->commit;
    return $message_id;
};



1;
