package Newslash::Plugin::AccessLog;
use Mojo::Base 'Mojolicious::Plugin';

use POSIX qw(strftime locale_h);
use IO::File;
use File::Basename;
use Mojo::Util qw(dumper);

sub handle {
    my $self = shift;
    my $fh;
    if ($self->{config}->{local_file}) {
        $fh = $self->_open_or_create_file($self->{config}->{local_file});
    }
    return $fh if $fh;
    return \*STDERR;
}

sub register {
    my ($self, $app, $conf) = @_;
    $conf ||= {};

    # set default config value
    my $log_cnf = $app->config->{Log} ||= {};
    my $cnf = $log_cnf->{access_log} ||= {};

    $cnf->{local_file} ||= "/var/log/newslash/access.log";
    $cnf->{autoflush} = 1 if !defined $cnf->{autoflush};
    $cnf->{mode} = "console" if !$cnf->{mode};

    $cnf = {%$cnf, %$conf};
    $self->{app} = $app;
    $self->{config} = $cnf;

    return if $cnf->{mode} eq "console";
    my $enable_log = 0;

    # local file output
    if ($cnf->{mode} eq "local_file" && $cnf->{local_file}) {
        my $fh = $self->handle;
	if (!$fh) {
	    $app->log->error("AccessLog: cannot open access log file: $cnf->{local_file}");
	}
	$enable_log = 1;
    }
    else {
	$app->log->error("AccessLog: mode '$cnf->{mode}' is invalid. allowed mode: 'console', 'local_file'");
    }

    if ($enable_log) {
	$app->hook(after_build_tx => sub {
		       my ($tx, $app) = @_;
		       $tx->on(finish => sub {
				   my $tx = shift;
				   my $log = $self->_format_log($tx);
				   $self->write_log($log);
			       });
		   });
    }
}

sub write_log {
    my ($self, $log) = @_;
    if ($self->{config}->{local_file}) {
        $self->handle->print("$log\n");
    }
    if ($self->{config}->{mojo_log}) {
	$self->{app}->log->debug($log);
    }
}

sub _format_log {
    my ($self, $tx) = @_;

    # LogFormat "%h %l %u %t \"%r\" %>s %b \"%{Referer}i\" \"%{User-agent}i\"" combined
    # remote_host logname remote_user [DD/MON/YYYY:HH:MM:SS +tzone] "REQUEST" final_status size_of_response "referer" "user-agent"

    my $remote_host = $tx->original_remote_address || "-";
    my $logname = "-";
    my $remote_user = "-";

    my $locale_org = setlocale(LC_TIME, "C");
    my $timestamp = strftime('[%d/%b/%Y:%H:%M:%S %z]', localtime);
    setlocale(LC_TIME, $locale_org);

    # build request line
    my $url  = $tx->req->url;
    my $path = $url->path_query;
    my $method = $tx->req->method;
    my $version = $tx->req->version;
    my $request = "\"$method $path HTTP/$version\"";

    my $status = $tx->res->code || "-";
    my $size = $tx->res->body_size || "-";

    my $referer = $tx->req->headers->referrer;
    my $ua = $tx->req->headers->user_agent;

    $referer = $referer ? "\"$referer\"" : "-";
    $ua = $ua ? "\"$ua\"" : "-";

    return "$remote_host $logname $remote_user $timestamp $request $status $size $referer $ua";
}

sub _open_or_create_file {
    my ($self, $pathname) = @_;

    if (-w $pathname) {
        return IO::File->new($pathname, "a");
    }

    if (-e $pathname) {
        $self->{app}->log->error("AccessLog::LocalFile: $pathname exists, but not writable");
        return;
    }

    my $parent = dirname($pathname);
    if (! -d $parent) {
        $self->{app}->log->error("AccessLog::LocalFile: cannot create $pathname: $parent not exists");
        return;
    }

    my $fh = IO::File->new($pathname, "a");
    if (!$fh) {
        $self->{app}->log->error("AccessLog::LocalFile: cannot create $pathname");
        return;
    }

    return $fh;
}

1;
