package Newslash::Plugin::EasyCache;
use Mojo::Base 'Mojolicious::Plugin';
use Mojo::Log;

sub kvs { return shift->{kvs}; }
sub inmemory { return shift->{inmemory}; }
sub app { return shift->{app}; }

sub register {
    my ($self, $app, $conf) = @_;

    my $kvs = $app->kvs;
    $self->{app} = $app;
    $self->{kvs} = $kvs;
    $self->{inmemory} = $app->cache;

    if (!$self->{app}) {
        Mojo::Log->new->warn('EasyCache: no app given');
    }
    if (!$self->{kvs}) {
        Mojo::Log->new->warn('EasyCache: no kvs given');
    }

    $app->helper(ezcache => sub { state $ezcache = $self });
}

sub model {
    my ($self, $model_name, $options) = @_;
    $options ||= {};
    $options->{kvs} ||= $self->kvs;
    $options->{inmemory} ||= $self->inmemory;
    $options->{key_prefix} ||= $model_name;
    $options->{expire} ||= 300; # 300 sec
    $options->{app} = $self->app;

    my $model = $self->{app}->model($model_name);
    if (!$model) {
        $self->app->log->warn("EasyCache: invalid model name: $model_name");
        return;
    }
    return Newslash::Plugin::EasyCache::ModelWrapper->new($model, $options);
}

sub cached_read {
    my ($self, $key, $config, $func, @params) = @_;

    if (!ref($config)) {
        $config = { expire => $config };
    }
    my $mem_expire = $config->{memory_expire} || $config->{expire} || 30;
    my $kvs_expire = $config->{kvs_expire} || $config->{expire} || 30;

    if (!$self->inmemory && !$self->kvs) {
        # cache system is not available
        $self->app->log->warn("EasyCache: no cache system available");
        return $func(@params);
    }

    # at first, check in-memory cache
    if ($self->inmemory) {
        my $result = $self->inmemory->get($key);
        if (defined $result) {
            $self->app->log->debug("EasyCache: use inmemory cache for $key");
            return $result;
        }
    }
    else {
        $self->app->log->warn("EasyCache: inmemory cache not available");
    }

    # next, check KVS cache
    if ($self->kvs) {
        my $result = $self->kvs->get($key);
        if (defined $result) {
            $self->app->log->debug("EasyCache: use kvs cache for $key");

            # update in-memory cache
            if ($self->inmemory) {
                my $ttl = $self->kvs->ttl($key);
                if ($ttl > 0) {
                    $self->inmemory->set($key, $result, $ttl);
                }
            }
            return $result;
        }
    }
    else {
        $self->app->log->warn("EasyCache: KVS cache not available");
    }

    # finally, execute function
    $self->app->log->debug("EasyCache: no cache available for $key");
    my $result = $func->(@params);
    if (defined $result) {
        $self->app->log->debug("EasyCache: update kvs and inmemory cache for $key, expire: $mem_expire / $kvs_expire");
        $self->kvs->set($key, $result, $kvs_expire);
        $self->inmemory->set($key, $result, $mem_expire);
    }

    return $result;

}

package Newslash::Plugin::EasyCache::ModelWrapper;
use Mojo::Util qw(md5_sum dumper);
use Storable;
use Data::Dumper;

sub new {
    my ($class, $model, $options) = @_;
    return if !$model;
    $options ||= {};
    $options->{expire} ||= 300;

    bless {options => $options,
           _model => $model,
          }, $class;
}

sub key_prefix { return shift->{options}->{key_prefix}; }
sub model { return shift->{_model}; }
sub app { return shift->{options}->{app}; }

sub _param_to_hash {
    my $self = shift;
    my @params = @_;

    my $serialized = Storable::freeze([@params]);
    return md5_sum($serialized);
}


sub select {
    my ($self, @params) = @_;

    # get key to access cached data
    my $key = $self->key_prefix . ":" . $self->_param_to_hash(@params);
    #$self->app->log->debug(Dumper(@params));

    if ($self->{options}->{no_cache}) {
        return $self->model->select(@params);
    }

    return $self->app->ezcache->cached_read($key,
                                            $self->{options}->{expire},
                                            sub { $self->model->select(@_); },
                                            @params);
}

1;
