package Newslash::Plugin::Stats;
use Mojo::Base 'Mojolicious::Plugin';
use Data::Dumper;

use constant KEY_PREFIX => "newslash_stats:";
use constant START_TIME_FIELD => "start_time";
use constant REQUEST_COUNTER_KEY => "req_count";
use constant STORY_COUNTER_KEY => "story_count";
use constant EVENT_COUNTER_KEY => "event_count";

sub register {
    my ($self, $app, $conf) = @_;
    $conf ||= {};
    $self->{app} = $app;
    $self->{kvs} = $app->kvs;
    $self->{options} = $conf->{options} || {};

    $app->hook(after_dispatch=> sub {
                   my $c = shift;
                   my $method = $c->req->method;
                   my $code = $c->res->code;
                   if (!$method || !$code) {
                       $app->log->error("Plugin::Stats: update_request_counter error. method: $method, code: $code");
                       return;
                   }
                   my $rs = $self->update_request_counter($method, $code);
                   if (!defined $rs) {
                       $app->log->error("Plugin::Stats: update_request_counter error. method: $method, code: $code");
                   }
                   else {
                       #$app->log->info("Plugin::Stats: update_request_counter done. method: $method, code: $code, rs: $rs");
                   }
               });

    $app->helper(stats => sub { state $stats = $self });
}

sub update_request_counter {
    my ($self, $method, $code) = @_;
    my $key = KEY_PREFIX . REQUEST_COUNTER_KEY;
    my $field = uc($method) . ":$code";
    return if !$field || !$key;

    my $kvs = $self->{kvs};
    if (!$kvs->hexists($key, START_TIME_FIELD)) {
        my $time = $kvs->time;
        $kvs->hset($key, START_TIME_FIELD, $time->[0], scalar => 1);
    }
    return $kvs->hincrby($key, $field, 1);
}

sub increment_story_pageview_counter {
    my ($self, $stoid) = @_;
    my $key = KEY_PREFIX . STORY_COUNTER_KEY;
    my $field = $stoid;
    return if !$field || !$key;

    my $kvs = $self->{kvs};
    if (!$kvs->hexists($key, START_TIME_FIELD)) {
        my $time = $kvs->time;
        $kvs->hset($key, START_TIME_FIELD, $time->[0], scalar => 1);
    }
    return $kvs->hincrby($key, $field, 1);
}

sub aggregate_story_pageview_counters {
    my ($self) = @_;
    my $key = KEY_PREFIX . STORY_COUNTER_KEY;
    my $kvs = $self->{kvs};
    my $result = {};

    # get keys
    my $stoids = $kvs->hkeys($key);
    my $stories = $self->{app}->model('stories');
    my $error = 0;

    for my $stoid (@$stoids) {
        next if $stoid !~ m/^\d+$/;
        my $count = $kvs->hget($key, $stoid, scalar => 1);
        $self->{app}->log->debug("Stats::aggregate_story_pageview_counters: ${key}.[$stoid] - $count");

        if ($count !~ m/^\d+$/) {
            $kvs->hdel($key, $stoid);
            next;
        }

        my $rs;
        if ($count) {
            $rs = $stories->update2(stoid => $stoid, hits => { add => $count + 0});
        }

        if (!defined $rs) {
            $self->{app}->log->error("Stats::aggregate_story_pageview_counters: update stories failed. stoid: $stoid");
            $error = 1;
        }
        else {
            # reset count
            $kvs->hincrby($key, $stoid, -$count);
            if (!defined $kvs->atomic_hdel($key, $stoid, 0)) {
                $self->{app}->log->error("Stats::aggregate_story_pageview_counters: reset counter failed. stoid: $stoid")
            }
        }
    }

    # update start_time
    if (!$error) {
        my $time = $kvs->time;
        $kvs->hset($key, START_TIME_FIELD, $time->[0], scalar => 1);
    }

    return 1;
}

sub add_event_counter {
    my ($self, $event) = @_;
    my $key = KEY_PREFIX . EVENT_COUNTER_KEY;
    my $field = $event;
    return if !$field || !$key;

    my $kvs = $self->{kvs};
    if (!$kvs->hexists($key, START_TIME_FIELD)) {
        my $time = $kvs->time;
        $kvs->hset($key, START_TIME_FIELD, $time->[0], scalar => 1);
    }
    return $kvs->hincrby($key, $field, 1);
}

1;
