package Newslash::Plugin::Stories;
use Mojo::Base 'Mojolicious::Plugin';

use DateTime;
use DateTime::Format::MySQL;
use DateTime::Format::ISO8601;

has 'last_error';
has 'app';
has 'c';

sub register {
    my ($self, $app, $conf) = @_;
    $self->app($app);
    $app->helper(stories => sub { $self->c(shift); state $stories = $self; });

    # default config values
    my $cnf = $app->config->{Stories} ||= {};
    $cnf->{title_max_byte} ||= 100;
    $cnf->{submission_karma_bonus} = 3 if !defined $cnf->{submission_karma_bonus};
}

sub related_stories {
    my ($self, $stoid) = @_;
    my $stories = $self->app->model('stories');

    return $stories->get_related_items(stoid => $stoid);
}

sub latest {
    my $self = shift;
    my $args = {@_};
    my $c = $self->c;

    my $user = $c->stash('user');
    my $limit = $args->{limit} || $c->_get_config("item_per_page", 10);
    my $show_future = $user->{is_admin} ? 1 : 0;
    my $show_nonpublic = $user->{is_admin} ? 1 : 0;

    my $stories;
    if ($user->{is_login}) {
        $stories = $c->model('stories');
    }
    else {
        # TODO: check expire time
        $stories = $c->ezcache->model('stories', { expire => 60 });
    }
    my $items = $stories->select(hide_future => !$user->{is_admin},
                                 public_only => !$user->{is_admin},
                                 limit => $limit + 1,
                                 order_by => {create_time => 'desc'},
                                );

    return $items;
}

sub create {
    my ($self, @params) = @_;
    my $params = {@params};
    my $err = "";
    my $config = $self->app->config->{Stories} || {};
    my $stories = $self->model("stories");

    # check parameters
    # required parameters are: uid, title, introtext
    my $intro_text = $params->{introtext} || $params->{intro_text};
    $err = "no_introtext_given" if !$intro_text;
    $err = "no_title_given" if !$params->{title};
    $err = "no_uid_given" if !$params->{uid};

    # check title length
    if (length($params->{title}) > $config->{title_max_byte}) {
        $err = "title_is_too_long";
    }

    # check timestamp if exists. use ISO8601 style timestamp like: 2006-08-14T02:34:56-0600
    if ($params->{time}) {
        $params->{create_time} = $params->{time};
        delete $params->{time};
    }
    if ($params->{create_time}) {
        my $rex_timestamp = qr/
                                  ^(\d+)-(\d+)-(\d+)[^ 0-9]+(\d+):(\d+):(\d+(?:\.\d+)?)   # datetime
                                  (?:Z|([+-])(\d+):(\d+))?$                          # tz
                              /xi;
        if ($params->{create_time} =~ $rex_timestamp) {
            my $dt = DateTime::Format::ISO8601->parse_datetime($params->{time});
            $params->{create_time} = DateTime::Format::MySQL->format_datetime($dt);
        }
    }

    if ($err) {
        $self->last_error = $err;
        return;
    }

    my $dbh = $stories->start_transaction;

    # allocate sid (insert blank story)
    my ($sid, $stoid) = $stories->allocate_sid(base_datetime => $params->{create_time});
    if (!$sid) {
        $self->last_error = "sid allocation failed";
        $stories->rollback;
        return;
    }

    # insert into story_text
    my $rs = $stories->text->create(stoid => $stoid,
                                    title => $params->{title},
                                    introtext => $intro_text,
                                    bodytext => $params->{bodytext} || $params->{body_text});
    if (!$rs) {
        $self->last_error = "insert text failed";
        $stories->rollback;
        return;
    }

    # insert into story
    $rs = $self->update(stoid => $stoid, %$params);
    if (!$rs) {
        $self->last_error = "insert story failed";
        $stories->rollback;
        return;
    }

    # create discussion
    my $users = $stories->new_instance_of("Users");
    my $firehose = $stories->new_instance_of("Firehose");
    my $submissions = $stories->new_instance_of("Submissions");

    # transfer from journal/firehose discussion

    # create firehose item

    # update submitter info
    my $submission_id = $params->{submission_id} || $params->{subid};
    if ($submission_id) {
        # if the story is derived from submission, do some action
        # first, update firehose item correspod to submission
        my $rs = $firehose->update(type => "submission",
                                   src_id => $submission_id,
                                   accepted => 'yes',
                                  );

        # get submitter
        my $submission = $submissions->select(id => $submission_id);
        my $submitter_uid = $submission->{uid};
        if (!$users->is_anonymous_uid($submitter_uid)) {
            # update submitter's karma
            my $karma_bonus = $config->{submission_karma_bonus};
            if ($karma_bonus) {
                $users->update_karma(uid => $submitter_uid,
                                     add => $karma_bonus);
            };
            $users->class->clear_rookie(uid => $submitter_uid);
            $users->class->validate_submitter(uid => $submitter_uid);
        }
    }


}

1;

=encoding utf8

=head1 NAME

Newslash::Plugin::Stories - Newslash stories manipulation plugin

=head1 SYNOPSIS

  # Mojolicious
  $app->plugin('Newslash::Plugin::Stories');

=head1 DESCRIPTION

L<Newslash::Plugin::Stories> is 'Business Logic' layer of Newslash.


=head1 HELPERS

L<Mojolicious::Plugin::Stories> implements the following helpers.

=head2 boxes

  [% helpers.boxes() %]

Fetch box contents for current user and returns them.

=head2 format_timestamp

  $c->format_timestamp(user => $user, epoch => $epoch, format => "user")
  $c->format_timestamp(datetime => $dt, format => "simple")

Return formated string.

=head1 METHODS

=head2 register

  $plugin->register(Mojolicious->new);

Register helpers in L<Mojolicious> application.

=head1 SEE ALSO

L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.

=cut
