package Newslash::Plugin::Users;
use Mojo::Base 'Mojolicious::Plugin';
use Mojo::JSON qw(to_json);
use Email::Valid;
use DateTime;

has 'last_error';
has 'app';

sub register {
    my ($self, $app, $conf) = @_;
    $self->app($app);
    $app->helper(users => sub { state $users = $self; });

    # default config values
    my $cnf = $app->config->{Users} ||= {};
    $cnf->{newpasswd_expiration} ||= 60 * 60 * 24; # 60[sec] * 60[min] * 24[hour]
    $cnf->{update_email_expiration} ||= 60 * 60 * 24; # 60[sec] * 60[min] * 24[hour]
}

sub get_user {
    my $self = shift;
    my $args = {@_};
    my $users = $self->app->model('users');

    my $user = $users->select(@_);
    return if !$user;

    # add .config
    $users->get_user_config($user);
    my $converted = $users->prefs->convert_old_prefs($user->{uid});
    if ($converted) {
        $user->{config} = $converted;
        $user->{configJSON} = to_json($converted);
    }

    # add .permissions
    $user->{permissions} = $users->permissions->get(user => $user) || {};

    return $user;

}

sub request_change_email {
    my ($self, $user, $email) = @_;

    if (!$user || !$user->{uid} || !$user->{is_login}) {
        $self->last_error("INVALID_USER");
        return;
    }

    if (!$email || !Email::Valid->address($email)) {
        $self->last_error("INVALID_ADDRESS");
        return;
    }

    # check if addres is already used
    my $users = $self->app->model('users');
    my $rs = $users->select(realemail => $email);
    #if (!defined $rs) {
    #    $self->app->log->error("Users: query error! email: $email");
    #    $self->last_error($user->last_error);
    #    return;
    #}
    if ($rs) {
        $self->last_error("ADDRESS_EXISTS");
        return;
    }

    # use users_param table to save temporary new email address
    my $param = $users->param;

    $rs = $param->set(uid => $user->{uid},
                      name => "new_email",
                      value => $email);

    if (!defined $rs) {
        $self->app->log->error("Users: update users_param table for request_change_email failed! uid: $user->{uid}");
        $self->last_error($param->last_error);
        return;
    }

    $self->app->event_que->emit("user", "change_email", $user->{uid}, $user->{uid}, 10);
    $self->app->stats->add_event_counter("change_email_request");
    return 1;
}

sub generate_email_change_token {
    my ($self, $uid) = @_;
    my $users = $self->app->model('users');

    # generate new token
    my $token = $users->passwords->generate_random_password;
    my $hashed_token = $users->passwords->encrypt_password($token);

    # insert token into users_param
    $users->param->start_transaction;
    my $rs = $users->param->set(uid => $uid,
                                name => "new_email_token",
                                value => $hashed_token);
    if (!defined $rs) {
        # error !
        $self->last_error($users->param->last_error);
        $users->param->rollback;
        return;
    }

    # insert token_ts intro users_param
    $rs = $users->param->set(uid => $uid,
                             name => "new_email_token_ts",
                             value => {function => "NOW()"});
    if (!defined $rs) {
        # error !
        $self->last_error($users->param->last_error);
        $users->param->rollback;
        return;
    }
    $users->param->commit;

    return $token;
}

sub verify_and_change_email {
    my ($self, $uid, $token) = @_;

    if (!$uid || !$token) {
        $self->last_error("INVALID_TOKEN");
        return;
    }

    my $users = $self->app->model('users');
    my $params = $users->param->select(uid => $uid);

    # check if token is expired
    my $token_ts = $params->{new_email_token_ts};
    if (!$token_ts) {
        $self->last_error("INVALID_TOKEN");
        return;
    }
    my $expire_dt = eval { DateTime::Format::MySQL->parse_datetime($token_ts) };
    my $expiration_limit = $self->app->config->{Users}->{update_email_expiration};
    if (!$expire_dt) {
        $self->app->log->error("Users: invalid new_email_token_ts ($token_ts). uid: $uid");
        $self->last_error("INVALID_TOKEN");
        return;
    }
    $expire_dt->add( seconds => $expiration_limit);
    if ($expire_dt->epoch() < time()) {
        $self->last_error("TOKEN_EXPIRED");
        return;
    }

    # check if token is valid
    if (!$users->passwords->verify_password(password => $token,
                                            hashed => $params->{new_email_token})) {
        $self->last_error("INVALID_TOKEN");
        return;
    }

    # check if address is valid
    my $email = $params->{new_email};
    if (!$email) {
        $self->last_error("INVALID_ADDRESS");
        return;
    }

    # check if address is already used
    my $rs = $users->select(realemail => $email);

    #if (!defined $rs) {
    #    $self->app->log->error("Users: query error! email: $email");
    #    $self->last_error($users->last_error || "QUERY_ERROR");
    #    return;
    #}
    if ($rs) {
        $self->last_error("ADDRESS_EXISTS");
        return;
    }

    # all OK, change realemail!
    $users->start_transaction;
    if (!defined $users->update(uid => $uid, realemail => $email)) {
        $users->rollback;
        $self->last_error($users->last_error || "UPDATE_ERROR");
        return;
    }
    if (!defined $users->param->delete(uid => $uid, name => "new_email")) {
        $users->rollback;
        $self->last_error($users->last_error || "DELETE_EMAIL_ERROR");
        return;
    }
    if (!defined $users->param->delete(uid => $uid, name => "new_email_token")) {
        $users->rollback;
        $self->last_error($users->last_error || "DELETE_TOKEN_ERROR");
        return;
    }
    if (!defined $users->param->delete(uid => $uid, name => "new_email_token_ts")) {
        $users->rollback;
        $self->last_error($users->last_error || "DELETE_TS_ERROR");
        return;
    }
    $users->commit;
    $self->app->stats->add_event_counter("change_email");
    return 1;
}

sub cancel_change_email_request {
    my ($self, $uid) = @_;
    if (!$uid) {
        $self->last_error("INVALID_UID");
        return;
    }
    my $users = $self->app->model('users');
    my $params = $users->param->select(uid => $uid);

    my $email = $params->{new_email};
    if (!$email) {
        $self->last_error("NO_ADDRESS");
        return;
    }

    $users->start_transaction;
    if (!defined $users->param->delete(uid => $uid, name => "new_email")) {
        $users->rollback;
        $self->last_error($users->last_error);
        return;
    }
    if (!defined $users->param->delete(uid => $uid, name => "new_email_token")) {
        $users->rollback;
        $self->last_error($users->last_error);
        return;
    }
    if (!defined $users->param->delete(uid => $uid, name => "new_email_token_ts")) {
        $users->rollback;
        $self->last_error($users->last_error);
        return;
    }
    $users->commit;
    $self->app->stats->add_event_counter("change_email_cancel");
    return 1;

}

sub reset_password {
    my ($self, $user) = @_;

    if (!$user || !$user->{uid} || !$user->{is_login}) {
        $self->last_error("INVALID_USER");
        return;
    }

    $self->app->event_que->emit("user", "reset_password", $user->{uid}, $user->{uid}, 10);
    $self->app->stats->add_event_counter("reset_password_request");
    return 1;
}

sub generate_password_reset_token {
    my ($self, $uid) = @_;
    my $users = $self->app->model('users');

    # generate new token
    my $token = $users->passwords->generate_random_password;
    my $hashed_token = $users->passwords->encrypt_password($token);

    my $rs = $users->update(uid => $uid,
                            newpasswd => $hashed_token,
                            newpasswd_ts => { function => "NOW()" });
    if (!$rs) {
        $self->last_error($users->last_error);
        return;
    }

    return $token;
}

sub cancel_activation {
    my ($self, $user) = @_;
    if (!$user || !$user->{uid}) {
        $self->last_error("INVALID_USER");
        return;
    }

    my $users = $self->app->model('users');

    my $rs = $users->update(uid => $user->{uid},
                            newpasswd => "",
                            newpasswd_ts => {function => "NULL"});
    if (!defined $rs) {
        $self->app->log->error("Users: newpasswd reset error! uid: $user->{uid}");
        $self->last_error($users->last_error);
        return;
    }
    $self->app->stats->add_event_counter("user_activation_cancel");
    return 1;
}

sub update_password_by_token {
    my ($self, $nickname, $token, $password) = @_;

    # check nickname and token pair
    my $user = $self->check_activation_token($nickname, $token);
    return if !$user;

    return $self->_update_password($user, $password);
}

sub update_password {
    my ($self, $user, $old_password, $new_password) = @_;
    if (!$user || !$user->{uid}) {
        $self->last_error("INVALID_USER");
        return;
    }

    my $users = $self->app->model('users');

    if ($users->passwords->verify_password(uid => $user->{uid},
                                           password => $old_password)) {
        # old password is correct.
        return $self->_update_password($user, $new_password);
    }

    # old password is incorrect!
    $self->last_error("INCORRECT_PASSWORD");
    return;
}

sub _update_password {
    my ($self, $user, $password) = @_;

    if (!$user || !$user->{uid}) {
        $self->last_error("INVALID_USER");
        return;
    }

    my $users = $self->app->model('users');

    my @params = (uid => $user->{uid},
                  passwd => $password );

    if ($user->{seclev} < 1) {
        push @params, seclev => 1;
    }

    if ($user->{newpasswd}) {
        push @params, newpasswd => "";
        push @params, newpasswd_ts => { function => "NULL" };
    }

    my $rs = $users->update(@params);

    if (!defined $rs) {
        $self->last_error($users->last_error);
        return;
    }
    $self->app->event_que->emit("user", "update_password", $user->{uid}, $user->{uid}, 10);
    $self->app->stats->add_event_counter("update_password");
    return 1;
}

sub check_activation_token {
    my ($self, $nickname, $token) = @_;
    return if (!$nickname || !$token);

    my $users = $self->app->model('users');
    my $user = $users->select(nickname => $nickname);

    # check if token is correct
    if (!$user
        || !$users->passwords->compare_password($token, $user->{newpasswd})
        || !$user->{newpasswd_ts}) {
        $self->last_error("INVALID_TOKEN");
        return;
    }

    # check if token is expired
    my $expiration_limit = $self->app->config->{Users}->{newpasswd_expiration};
    my $expire_dt = eval { DateTime::Format::MySQL->parse_datetime($user->{newpasswd_ts}) };
    if (!$expire_dt) {
        $self->app->log->error("Users: invalid newpasswd_ts ($user->{newpasswd_ts}). uid: $user->{uid}");
        $self->last_error("INVALID_TOKEN");
        return;
    }
    $expire_dt->add( seconds => $expiration_limit);
    if ($expire_dt->epoch() < time()) {
        $self->last_error("TOKEN_EXPIRED");
        return;
    }

    # ok
    return $user;
}

sub create_new_user {
    my ($self, $nickname, $email, $options) = @_;
    my $users = $self->app->model('users');
    $options ||= {};

    # check $nickname and $email
    my ($id_error, $email_error) = (0, 0);
    $id_error = $self->validate_nickname($nickname) ? 0 : $self->last_error;
    $email_error = $self->validate_email($email) ? 0 : $self->last_error;

    if ($id_error || $email_error) {
        $self->last_error({ id_error => $id_error,
                            email_error => $email_error });
        return;
    }

    my $uid = $users->create($nickname, $email, "", { seclev => 0 });
    if (!$uid) {
        # error occured
        $self->last_error({ id_error => $id_error,
                            email_error => $email_error,
                            system_error => $users->last_error });
        return;
    }

    # check options
    if ($options->{message}) {
        my $message_types = $self->app->model('messages');
        for my $k (keys %{$options->{message}}) {
            my $rs = $users->messages->update(uid => $uid,
                                              name => $k,
                                              mode => $options->{message}->{$k});
            if (!defined $rs) {
                $self->app->log->error("Users: message update failed! uid: $uid, name: $k, mode: $options->{message}->{$k}");
            }
        }
    }

    $self->app->event_que->emit("user", "create", $uid, $uid, 10);
    $self->app->stats->add_event_counter("user_create");
    return $uid;
}

sub validate_nickname {
    my ($self, $nickname) = @_;
    my $nick_regex = qr/^[a-zA-Z_][ a-zA-Z0-9\$_.+!*\'(),-]{0,19}$/;
    my $users = $self->app->model('users');

    if (!$nickname) {
        $self->last_error("BLANK_ID");
        return;
    }

    if ($nickname !~ $nick_regex) {
        $self->last_error("INVALID_ID");
        return;
    }

    my $matchname = $users->nickname_to_matchname($nickname);
    my $rs = $users->select(matchname => $matchname);
    if ($rs) {
        $self->last_error("ID_EXISTS");
        return;
    }

    return 1;
}

sub validate_email {
    my ($self, $email) = @_;
    my $users = $self->app->model('users');

    if (!$email) {
        $self->last_error("BLANK_ADDRESS");
        return;
    }

    if (!Email::Valid->address($email)) {
        $self->last_error("INVALID_ADDRESS");
        return;
    }

    my $rs = $users->select(realemail => $email);
    if ($rs) {
        $self->last_error("ADDRESS_EXISTS");
        return;
    }

    return 1;
}


1;

=encoding utf8

=head1 NAME

Newslash::Plugin::Users - Newslash users manipulation plugin

=head1 SYNOPSIS

  # Mojolicious
  $app->plugin('Newslash::Plugin::Users');

=head1 DESCRIPTION

L<Newslash::Plugin::Users> is 'Business Logic' layer of Newslash.


=head1 HELPERS

L<Mojolicious::Plugin::Users> implements the following helpers.

=head2 users

  my $u = helpers.users();

Returns Newslash::Plugin::Users instance.

=head1 METHODS

=head2 register

  $plugin->register(Mojolicious->new);

Register helpers in L<Mojolicious> application.

=head2 request_change_email

=head2 generate_email_change_token

=head2 verify_and_change_email

=head2 cancel_change_email_request

=head2 rest_password

=head2 generate_password_reset_token

=head2 cancel_activation

=head2 update_password_by_token

=head2 update_password

=head2 activation

=head2 create_new_user

=head2 validate_nickname

=head2 validate_email

=head1 SEE ALSO

L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.

=cut
