package Newslash::Web::Controller::Archive;
use Mojo::Base 'Mojolicious::Controller';
use Mojo::Util qw(dumper);
use Mojo::JSON qw(decode_json encode_json);
use Template;

use DateTime;
use DateTime::Format::MySQL;
use Newslash::Util::Items;

sub _check_params {
    my ($c, $type, $unit) = @_;

    my $year = $c->stash('year');
    my $month = $c->stash('month');
    my $day = $c->stash('day');
    my $do_redirect = 0;
    my $r_path = "";

    # check year
    if (!$year || $year !~ m/^\d{2}$/) {
        $year = DateTime->now->year - 2000;
        $month = DateTime->now->month;
        $month = "0" . $month if ($month < 10);
        $day = "01";
        $do_redirect = 1;
    }
    # check month
    if (!$month || $month !~ m/^\d{2}$/) {
        $month = "01";
        $day = "01";
        if ($unit eq "month" || $unit eq "day") {
            $do_redirect = 1;
        }
    }
    # check day
    if (!$day || $day !~ m/^\d{2}$/) {
        $day = "01";
        if ($unit eq "day") {
            $do_redirect = 1;
        }
    }

    if ($unit eq "year") {
        $r_path = "/$type/$year/" if $do_redirect;
        return ($year, $month, $day, $r_path);
    }
    if ($unit eq "month") {
        $r_path = "/$type/$year/$month/" if $do_redirect;
        return ($year, $month, $day, $r_path);
    }
    if ($unit eq "day") {
        $r_path = "/$type/$year/$month/$day/" if $do_redirect;
        return ($year, $month, $day, $r_path);
    }

    # invalid unit
    $year = DateTime->now->year - 2000;
    $month = DateTime->now->month;
    $month = "0" . $month if ($month < 10);
    $day = "01";
    $r_path = "/$type/$year/$month/$day/";
    return ($year, $month, $day, $r_path);
}

sub _render_archive {
    my $c = shift;
    my $params = {@_};

    my $user = $c->stash('user');
    my $model = $params->{model};
    my $type = $params->{type};
    my $unit = $params->{unit} || "day";

    my ($year, $month, $day, $r_path) = _check_params($c, $type, $unit);

    if ($r_path) {
        $c->res->code(301);
        $c->redirect_to($r_path);
        return;
    }

    $year = $year + 2000;
    my $offset_sec = $user->{config}->{ui}->{offset_sec} || 0;
    my $q_params = { year => $year,
                     month => $month,
                     day => $day,
                     offset_sec => $offset_sec,
                     order_by => {create_time => "DESC"},
                   };

    my $unwanted_keys = [];
    $unwanted_keys = [qw(month day)] if ($unit eq "year");
    $unwanted_keys = [qw(day)] if ($unit eq "month");
    for my $k (@$unwanted_keys) {
        delete $q_params->{$k};
    }

    if (!$user->{author}) {
        $q_params->{hide_future} = 1;
        $q_params->{public_only} = 1;
    }

    my $items = $model->select(%$q_params);

    delete $q_params->{$unit};
    delete $q_params->{order_by};
    my $counts = $model->count(%$q_params);

    # get next/prev date
    my $next;
    my $prev;
    my $dt = DateTime->new(year => $year,
                           month => $month,
                           day => $day);
    my $prev_items = $model->select(until => $dt,
                                    offset_sec => $offset_sec,
                                    order_by => {create_time => "DESC"},
                                    hide_future => $user->{author} ? 0 : 1,
                                    public_only => $user->{author} ? 0 : 1,
                                    limit => 1,
                                   );

    $dt->add(days => 1) if $unit eq "day";
    $dt->add(months => 1) if $unit eq "month";
    my $next_items = $model->select(since => $dt,
                                    offset_sec => $offset_sec,
                                    order_by => {create_time => "ASC"},
                                    hide_future => $user->{author} ? 0 : 1,
                                    public_only => $user->{author} ? 0 : 1,
                                    limit => 1
                                   );
    if ($prev_items && @$prev_items) {
        $prev = Newslash::Util::Items->get_date_string_from_item($prev_items->[0], $offset_sec, $unit);
    }

    if ($next_items && @$next_items) {
        $next = Newslash::Util::Items->get_date_string_from_item($next_items->[0], $offset_sec, $unit);
    }


    # get prev month
    $dt = DateTime->new(year => $year,
			month => $month,
			day => 1);
    my ($prev_month, $prev_month_date);
    $dt->add(days => -1);
    $prev_items = $model->select(until => $dt,
				 offset_sec => $offset_sec,
				 order_by => {create_time => "DESC"},
				 hide_future => $user->{author} ? 0 : 1,
				 public_only => $user->{author} ? 0 : 1,
				 limit => 1
				);
    if ($prev_items && @$prev_items) {
	$dt = $c->format_timestamp(mysql => $prev_items->[0]->{create_time},
				   format => "datetime",
				   offset_sec => 0,
				  );
	$dt->set(day => 1, hour => 0, minute => 0, second => 0);
	$prev_items = $model->select(since => $dt,
				     offset_sec => $offset_sec,
				     order_by => {create_time => "ASC"},
				     hide_future => $user->{author} ? 0 : 1,
				     public_only => $user->{author} ? 0 : 1,
				     limit => 1
				    );
	if ($prev_items && @$prev_items) {
	    $prev_month_date = Newslash::Util::Items->get_date_string_from_item($prev_items->[0], $offset_sec, "day");
	    $prev_month_date =~ m/^\d+\/0?(\d+)/;
	    $prev_month = $1;
	}
    }

    # get next month
    $dt = DateTime->new(year => $year,
			month => $month,
			day => 1);
    my ($next_month, $next_month_date);
    $dt->add(months => 1);
    $next_items = $model->select(since => $dt,
				 offset_sec => $offset_sec,
				 order_by => {create_time => "ASC"},
				 hide_future => $user->{author} ? 0 : 1,
				 public_only => $user->{author} ? 0 : 1,
				 limit => 1
				);
    if ($next_items && @$next_items) {
	$next_month_date = Newslash::Util::Items->get_date_string_from_item($next_items->[0], $offset_sec, "day");
	$next_month_date =~ m/^\d+\/0?(\d+)/;
	$next_month = $1;
    }

    my $page = { content_type => $type,
                 type => "archive",
                 year => $year,
                 month => $month,
                 day => $day,
               };

    $c->render(template => 'archive/index',
               items => $items,
               counts => $counts,
               page => $page,
               prev => $prev,
               next => $next,
	       prev_month => $prev_month,
	       next_month => $next_month,
	       prev_month_date => $prev_month_date,
	       next_month_date => $next_month_date,
	       no_ad => (@$items == 0),
              );

    if (@$items == 0) {
        $c->rendered(404);
    }
}

sub story {
    my $c = shift;
    $c->_render_archive(model => $c->model("stories"),
                        type => "story",
                        title => "記事");
}

sub journal {
    my $c = shift;
    $c->_render_archive(model => $c->model("journals"),
                        type => "journal",
                        title => "日記");
}

sub comment {
    my $c = shift;
    $c->_render_archive(model => $c->model("comments"),
                        type => "comment",
                        title => "コメント");
}

sub submission {
    my $c = shift;
    $c->_render_archive(model => $c->model("submissions"),
                        type => "submission",
                        title => "タレコミ");
}

sub poll {
    my $c = shift;
    $c->_render_archive(model => $c->model("polls"),
                        type => "poll",
                        title => "国民投票",
                        unit => "month");
}

1;
