# -*-Perl-*-
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Newslash::Util::TestMan;
use Data::Dumper;

use Newslash::Eventd::Handler::Comment;

my $t = Test::Mojo->new('Newslash::Web');

my ($admin, $user, $journal, $comment1, $comment2);

my $test_man = Newslash::Util::TestMan->new($t);
my $events = $t->app->model('events');
ok($events, "get Events instance");

my $comment_handler = new Newslash::Eventd::Handler::Comment($t->app);
ok($comment_handler, "create comment handler instance");

my $web_messages = $t->app->model('web_messages');
ok($web_messages, "get WebMessages instance");
my $messages = $t->app->model('messages');
ok($messages, "get Messages instance");

my $users = $t->app->model('users');
ok($users, "get Users instance");

my $test_comment_title = "test comment foobar";

# prepare for test
if ($t->app->mode eq 'test') {
    $admin = $test_man->create_admin("ehcadmin");
    ok($admin, "create admin user");

    $user = $test_man->create_user("ehctest");
    ok($user, "create test user");

    $journal = $test_man->create_journal(user => $admin);
    ok($journal, "create test journal");

    # set reply message's notification to web
    ok($users->messages->set(uid => $user->{uid},
                             code => 4,
                             mode => 1),
       "change notification mode for user");
    # check if update config correctly
    my $message_configs = $users->messages->select(uid => $user->{uid});
    my $mode = -1;
    for my $conf (@$message_configs) {
        $mode = $conf->{mode} if $conf->{code} == 4;
    }
    is($mode, 1, "update config correctly");
    diag Dumper $message_configs if $mode != 1;

    # set journal reply 's notification to web
    ok($users->messages->set(uid => $admin->{uid},
                             code => 7,
                             mode => 1),
       "change notification mode for admin");
    $message_configs = $users->messages->select(uid => $admin->{uid});
    $mode = -1;
    for my $conf (@$message_configs) {
        $mode = $conf->{mode} if $conf->{code} == 7;
    }
    is($mode, 1, "update config correctly");
    diag Dumper $message_configs if $mode != 1;

    $comment1 = $test_man->create_comment(user => $user,
                                          title => $test_comment_title,
                                          comment => "this is test comment",
                                          discussion_id => $journal->{discussion_id},
                                         );
    ok($comment1, "create comment1");

    $comment2 = $test_man->create_comment(user => $admin,
                                          title => "Re: " . $test_comment_title,
                                          comment => "this is test comment 2",
                                          discussion_id => $journal->{discussion_id},
                                          pid => $comment1->{cid},
                                         );
    ok($comment2, "create comment2");
}

subtest 'test comment:post handler' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    # create journal reply event
    $t->app->event_que->emit("comment", "post", $user->{uid}, $comment1->{cid});

    # get and process event
    my $event = $events->select(type => "comment",
                                subtype => "post",
                                related_id => $comment1->{cid},
                                active => 1,
                                order_by => { created_timestamp => "DESC" },
                                limit => 1,
                               );
    ok($event, "select events 1");
    ok($event->[0], "get valid event 1");
    if (!$event->[0]) {
        diag Dumper $event;
    }
    else {
        $event = $event->[0];
        $comment_handler->post($event);
        my $msgs = $web_messages->select(uid => $admin->{uid},
                                         order_by => { id => "DESC" },
                                         limit => 1);
        ok($msgs, "select web_message");
        is(@$msgs, 1, "select valid message");
        like($msgs->[0]->{subject}, qr/$journal->{title}/, "subject check 1-1");
        like($msgs->[0]->{subject}, qr/$comment1->{title}/, "subject check 1-2");
        like($msgs->[0]->{message}, qr/\/comment\/$comment1->{cid}/, "message check 1-1");

        ok($web_messages->hard_delete(id => $msgs->[0]->{id}), "delete message");
    }

    # create comment reply event
    # this event generate 2 messages
    $t->app->event_que->emit("comment", "post", $admin->{uid}, $comment2->{cid});

    # get and process event
    $event = $events->select(type => "comment",
                             subtype => "post",
                             related_id => $comment2->{cid},
                             active => 1,
                             order_by => { created_timestamp => "DESC" },
                             limit => 1,
                            );
    ok($event, "select events 2");
    ok($event->[0], "get valid event 2");
    if (!$event->[0]) {
        diag Dumper $event;
    }
    else {
        $event = $event->[0];
        $comment_handler->post($event);

        # check 1st message
        my $msgs = $web_messages->select(uid => $user->{uid},
                                         order_by => { id => "DESC" },
                                         limit => 1);
        ok($msgs, "select web_message 2");
        is(@$msgs, 1, "select valid message 2");
        like($msgs->[0]->{subject}, qr/$comment1->{title}/, "subject check 2-1");
        like($msgs->[0]->{subject}, qr/Re: /, "subject check 2-2");
        like($msgs->[0]->{message}, qr/\/comment\/$comment2->{cid}/, "message check 2-1");

        ok($web_messages->hard_delete(id => $msgs->[0]->{id}), "delete message 2");

        # check 2st message
        $msgs = $web_messages->select(uid => $admin->{uid},
                                      order_by => { id => "DESC" },
                                      limit => 1);
        ok($msgs, "select web_message 3");
        is(@$msgs, 1, "select valid message 3");
        like($msgs->[0]->{subject}, qr/$comment1->{title}/, "subject check 3-1");
        like($msgs->[0]->{subject}, qr/Re: /, "subject check 3-2");
        like($msgs->[0]->{message}, qr/\/comment\/$comment2->{cid}/, "message check 3-1");

        ok($web_messages->hard_delete(id => $msgs->[0]->{id}), "delete message 3");
    }


};

$test_man->cleanup;
done_testing;
