# -*-Perl-*-
# Newslash::Model::Relations related  tests
use Mojo::Base -strict;

use Test::More;
use Test::Mojo;
use Data::Dumper;
use Newslash::Util::Test qw(create_user delete_user
                            create_story delete_story
                            create_comments
                            create_users create_admin_user
                            get_anonymous_user
                          );

use constant {
      REL_FRIEND => "friends",
      REL_FREAK => "freaks",
      REL_FAN => "fans",
      REL_FOE => "foes",
      REL_FOF => "fofs",
      REL_EOF => "eofs",
  };

my $t = Test::Mojo->new('Newslash::Web');
my $users = $t->app->model('users');
my $relations = $t->app->model('relations');
ok($users, "get Users instance");
ok($relations, "get Relations instance");


my $test_users;

# prepare for test
if ($t->app->mode eq 'test') {
    # at first, create test users.
    $test_users = create_users($t->app, "reltest", 4);
}


subtest 'create/select/update relations' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    my $rs = $relations->create($test_users->{reltest1}->{uid},
                                $test_users->{reltest2}->{uid},
                                'friend');
    ok($rs, "relation create");
    diag($relations->last_error) if !$rs;

    my $friends = $relations->friends($test_users->{reltest1}->{uid});
    ok($friends, "get friends");
    is($friends->[0], $test_users->{reltest2}->{uid}, "get valid frineds");

    $rs = $relations->delete($test_users->{reltest1}->{uid},
                                $test_users->{reltest2}->{uid});
    ok($rs, "relation delete");

    $friends = $relations->friends($test_users->{reltest1}->{uid});
    ok($friends, "get friends (2)");
    is(@$friends, 0, "delete frineds succeed");
};


subtest 'get relations hash' => sub {
    plan skip_all => "mode is 'test'" if ($t->app->mode eq 'test');

    my $people = $relations->hash(4);
    ok($people, "retrive people");

    my $info = $users->select(target => 'info', uid => 4);
    ok($info, "retrive from users_info");

    my $people_raw = $info->{people};
    my $people_info = Storable::thaw($people_raw);
    ok($people_info, "nfreeze serialized people data");

    is_deeply($people, $people_info, "check people from people table and from users_info table");
};


subtest 'manipulate relations' => sub {
    plan skip_all => "mode is not 'test'" if ($t->app->mode ne 'test');

    # at first, create test users.
    my ($uid01, $uid02, $uid03, $uid04) = ($test_users->{reltest1}->{uid},
                                           $test_users->{reltest2}->{uid},
                                           $test_users->{reltest3}->{uid},
                                           $test_users->{reltest4}->{uid});

    my $rs = $relations->create($uid01, $uid02, "friend");
    ok($rs, "create relation 1");

    $rs = $relations->create($uid02, $uid03, "friend");
    ok($rs, "create relation 2");

    $rs = $relations->create($uid02, $uid04, "foe");
    ok($rs, "create relation 3");


    my $r1 = $relations->hash($uid01);
    is_deeply($r1, {
                    REL_FRIEND() => { $uid02 => 1, },
                    REL_FOF() => { $uid03 => { $uid02 => 1, }, },
                    REL_EOF() => { $uid04 => { $uid02 => 1, }, },
                   }, "reltest1's relations");

    my $friends = $relations->friends(uid => $uid01);
    is(@$friends, 1, "get friends");
    is($friends->[0], $uid02, "get valid friends");
    my $fofs = $relations->friends_of_friends(uid => $uid01);
    ok($fofs,  "get fofs");
    ok($fofs->{$uid03}, "get valid fofs");
    is($fofs->{$uid03}->[0], $uid02, "get valid fofs value");

    my $eofs = $relations->foes_of_friends(uid => $uid01);
    ok($eofs, "get eofs");
    ok($eofs->{$uid04}, "get valid eofs");
    is($eofs->{$uid04}->[0], $uid02, "get valid eofs value");

    ok($relations->is_friend(of => $uid01, uid => $uid02), "is_friend");
    ok(!$relations->is_friend(of => $uid01, uid => $uid03), "not is_friend");
    ok($relations->is_fof(of => $uid01, uid => $uid02), "is_fof");
    ok(!$relations->is_fof(of => $uid01, uid => $uid03), "not is_fof");
    ok($relations->is_eof(of => $uid01, uid => $uid02), "is_eof");
    ok(!$relations->is_eof(of => $uid01, uid => $uid03), "not is_eof");

    my $r2 = $relations->hash($uid02);
    is_deeply($r2, {
                    REL_FAN() => { $uid01 => 1, },
                    REL_FRIEND() => { $uid03 => 1, },
                    REL_FOE() => { $uid04 => 1, },
                   }, "reltest2's relations");
    my $foes = $relations->foes(uid => $uid02);
    is(@$foes, 1, "get foes");
    is($foes->[0], $uid04, "get valid foes");

    ok($relations->is_foe(of => $uid02, uid => $uid04), "is_friend");
    ok(!$relations->is_foe(of => $uid02, uid => $uid01), "not is_friend");

    my $r3 = $relations->hash($uid03);
    is_deeply($r3, {
                    REL_FAN() => { $uid02 => 1, },
                   }, "reltest3's relations");

    my $r4 = $relations->hash($uid04);
    is_deeply($r4, {
                    REL_FREAK() => { $uid02 => 1, },
                   }, "reltest4's relations");

};


# cleanup
if ($t->app->mode eq 'test') {
    for my $u (keys %$test_users) {
        delete_user($t->app, $u);
    }
}

done_testing();
