# coding: utf-8
'''submissions.py: newslash submissions related database module'''

import random
from base import NewslashDB, NewslashDBError

CREATE_POLLQUESTIONS_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `pollquestions` (
  `qid`         mediumint(8) unsigned               NOT NULL AUTO_INCREMENT,
  `question`    varchar(255)                        NOT NULL DEFAULT '',
  `voters`      mediumint(9)                        DEFAULT NULL,
  `topic`       int(10) unsigned                    NOT NULL,
  `discussion`  mediumint(8) unsigned               NOT NULL DEFAULT '0',
  `date`        datetime                            DEFAULT NULL,
  `uid`         mediumint(8) unsigned               NOT NULL DEFAULT '0',
  `primaryskid` smallint(5) unsigned                DEFAULT NULL,
  `autopoll`    enum('no','yes')                    NOT NULL DEFAULT 'no',
  `flags`       enum('ok','delete','dirty')         NOT NULL DEFAULT 'ok',
  `polltype`    enum('nodisplay','section','story') DEFAULT 'section',
  PRIMARY KEY             (`qid`),
  KEY `uid`               (`uid`),
  KEY `discussion`        (`discussion`),
  KEY `ibfk_converttid_2` (`topic`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

CREATE_POLLANSWERS_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `pollanswers` (
  `qid`    mediumint(9) NOT NULL DEFAULT '0',
  `aid`    mediumint(9) NOT NULL DEFAULT '0',
  `answer` char(255)    DEFAULT NULL,
  `votes`  mediumint(9) DEFAULT NULL,
  PRIMARY KEY (`qid`,`aid`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

class Polls(NewslashDB):
    def __init__(self, config):
        self.db = NewslashDB(config)

    def select(self, **kwargs):
        if not "limit" in kwargs:
            kwargs["limit"] = self.db.config("default_limit")

        if not "offset" in kwargs:
            kwargs["offset"] = 0

        if "qid" in kwargs:
            query = ("SELECT pollquestions.*, pollanswers.* FROM pollquestions"
                     "  LEFT JOIN pollanswers USING(qid)"
                     "  WHERE pollquestions.qid = %(qid)s"
                     "  LIMIT %(limit)s OFFSET %(offset)s")
        elif "date_ge" in kwargs:
            query = ("SELECT pollquestions.*, pollanswers.* FROM pollquestions"
                     "  LEFT JOIN pollanswers USING(qid)"
                     "  WHERE pollquestions.qid <= %(qid_le)s AND pollquestions.date >= %(date_ge)s"
                     "  ORDER BY pollquestions.qid ASC LIMIT %(limit)s OFFSET %(offset)s")
        elif "qid_gt" in kwargs:
            query = ("SELECT pollquestions.*, pollanswers.* FROM pollquestions"
                     "  LEFT JOIN pollanswers USING(qid)"
                     "  WHERE pollquestions.qid > %(qid_gt)s"
                     "  ORDER BY pollquestions.qid ASC LIMIT %(limit)s OFFSET %(offset)s")
        else:
            query = ("SELECT pollquestions.*, pollanswers.* FROM pollquestions"
                     "  LEFT JOIN pollanswers USING(qid)"
                     "  ORDER BY pollquestions.qid DESC LIMIT %(limit)s OFFSET %(offset)s")

        cur = self.db.execute(query, **kwargs)

        col_names = cur.column_names
        answers = {}
        result = []
        for item in cur:
            d = dict(zip(col_names, item))
            qid = d["qid"]
            if qid not in answers:
                answers[qid] = []
                d["answers"] = answers[qid]
                result.append(d)
            answers[qid].append(dict(qid=qid, aid=d["aid"], answer=d["answer"], votes=d["votes"]))

        for item in result:
            del item["aid"]
            del item["answer"]
            del item["votes"]
        
        self.db.close()
        return result

    def create_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi((CREATE_POLLANSWERS_TABLE_SQL,
                               CREATE_POLLQUESTIONS_TABLE_SQL))

    def drop_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi(('DROP TABLE pollanswers',
                               'DROP TABLE pollquestions'))

    def insert(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()
        answers = kwargs["answers"]
        del kwargs["answers"]
        try:
            cur = db.execute('INSERT INTO pollquestions (question, topic, date, uid)'
                             'VALUE (%(question)s, %(topic)s, NOW(), %(uid)s)',
                             **kwargs)
            kwargs["qid"] = cur.lastrowid
            aid = 1
            for answer in answers:
                args = dict(qid=kwargs["qid"], aid=aid, answer=answer)
                db.generic_insert("pollanswers", ("qid", "aid", "answer"), args)
                aid += 1
        except NewslashDBError as e:
            db.rollback()
            raise e

        db.commit()
        db.close()
        return kwargs["qid"]

    def update(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()
        answers = kwargs["answers"]
        del kwargs["answers"]

        try:
            db.generic_update('pollquestions', ('question', 'polltype', 'date'), ('qid',), kwargs)
            aid = 1
            for answer in answers:
                args = dict(qid=kwargs["qid"], aid=aid, answer=answer)
                db.generic_update('pollanswers', ('answer',), ('qid', 'aid'), args)
                aid += 1
        except NewslashDBError as e:
            db.rollback()
            raise e

        db.commit()
        db.close()
        return


    
