# coding: utf-8
'''submissions.py: newslash submissions related database module'''

import random
from base import NewslashDB, NewslashDBError

CREATE_SUBMISSIONS_TABLE_SQL = '''
CREATE TABLE IF NOT EXISTS `submissions` (
  `subid`       mediumint(8) unsigned                       NOT NULL AUTO_INCREMENT,
  `email`       varchar(255)                                NOT NULL DEFAULT '',
  `emaildomain` varchar(255)                                NOT NULL DEFAULT '',
  `name`        varchar(50)                                 NOT NULL DEFAULT '',
  `time`        datetime                                    NOT NULL DEFAULT '1970-01-01 00:00:00',
  `subj`        varchar(80)                                 NOT NULL,
  `story`       mediumtext                                  NOT NULL,
  `tid`         int(10) unsigned                            NOT NULL,
  `note`        varchar(30)                                 NOT NULL DEFAULT '',
  `primaryskid` smallint(5) unsigned                        DEFAULT NULL,
  `comment`     varchar(255)                                NOT NULL DEFAULT '',
  `uid`         mediumint(8) unsigned                       NOT NULL DEFAULT '0',
  `ipid`        varchar(32)                                 NOT NULL DEFAULT '',
  `subnetid`    varchar(32)                                 NOT NULL DEFAULT '',
  `del`         tinyint(4)                                  NOT NULL DEFAULT '0',
  `weight`      float                                       NOT NULL DEFAULT '0',
  `signature`   varchar(32)                                 NOT NULL DEFAULT '',
  `mediatype`   enum('none','text','video','image','audio') NOT NULL DEFAULT 'none',
  PRIMARY KEY            (`subid`),
  UNIQUE KEY             `signature` (`signature`),
  KEY `del`              (`del`),
  KEY `uid`              (`uid`),
  KEY `emaildomain`      (`emaildomain`),
  KEY `time_emaildomain` (`time`,`emaildomain`),
  KEY `ipid`             (`ipid`),
  KEY `subnetid`         (`subnetid`),
  KEY `primaryskid_tid`  (`primaryskid`,`tid`),
  KEY `tid`              (`tid`)
) ENGINE=InnoDB DEFAULT CHARSET=utf8mb4 ROW_FORMAT=DYNAMIC
'''

class Submissions(NewslashDB):
    def __init__(self, config):
        self.db = NewslashDB(config)

    def select(self, **kwargs):
        if not "limit" in kwargs:
            kwargs["limit"] = self.db.config("default_limit")

        if not "offset" in kwargs:
            kwargs["offset"] = 0

        if "subid" in kwargs:
            query = ("SELECT submissions.*, firehose.public, firehose.last_update FROM submissions"
                     "  LEFT JOIN firehose ON firehose.type = 'submission' AND firehose.srcid = submissions.subid"
                     "  WHERE submissions.subid = %(subid)s"
                     "  LIMIT %(limit)s OFFSET %(offset)s")
        elif "last_update_ge" in kwargs:
            query = ("SELECT submissions.*, firehose.public, firehose.last_update FROM submissions"
                     "  LEFT JOIN firehose ON firehose.type = 'submission' AND firehose.srcid = submissions.subid"
                     "  WHERE submissions.subid <= %(subid_le)s AND firehose.last_update >= %(last_update_ge)s"
                     "  ORDER BY submissions.subid ASC LIMIT %(limit)s OFFSET %(offset)s")
        elif "subid_gt" in kwargs:
            query = ("SELECT submissions.*, firehose.public, firehose.last_update FROM submissions"
                     "  LEFT JOIN firehose ON firehose.type = 'submission' AND firehose.srcid = submissions.subid"
                     "  WHERE submissions.subid > %(subid_gt)s"
                     "  ORDER BY submissions.subid ASC LIMIT %(limit)s OFFSET %(offset)s")
        else:
            query = ("SELECT submissions.*, firehose.public, firehose.last_update FROM submissions"
                     "  LEFT JOIN firehose ON firehose.type = 'submission' AND firehose.srcid = submissions.subid"
                     "  ORDER BY submissions.subid DESC LIMIT %(limit)s OFFSET %(offset)s")

        cur = self.db.execute(query, **kwargs)
        col_names = cur.column_names
        result = [dict(zip(col_names, x)) for x in cur]

        self.db.close()
        return result

    def create_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi((CREATE_SUBMISSIONS_TABLE_SQL,))

    def drop_tables(self):
        '''this function implemented for test purpose...'''
        self.db.execute_multi(('DROP TABLE submissions',))

    def insert(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()
        kwargs["globjid"] = random.randrange(1, 60000)
        kwargs["type"] = "submission";
        kwargs["name"] = "foobar";
        kwargs["sprite_info"] = "";
        try:
            cur = db.generic_insert("submissions", ("subj", "story", "uid", "tid", "signature"), kwargs)
            kwargs["subid"] = cur.lastrowid
            kwargs["srcid"] = cur.lastrowid
            db.generic_insert('firehose', ('type', 'globjid', 'public', 'name', 'sprite_info', 'srcid'), kwargs)
        except NewslashDBError as e:
            db.rollback()
            raise e


        db.commit()
        db.close()
        return kwargs["subid"]

    def update(self, **kwargs):
        '''this function implemented for test purpose...'''
        db = self.db.start_transaction()

        try:
            db.generic_update('submissions', ('subj', 'story'), ('subid',), kwargs)
            if 'public' in kwargs:
                kwargs["srcid"] = kwargs["subid"]
                kwargs["type"] = "submission"
                db.generic_update('firehose', ('public',), ('srcid', 'type'), kwargs)
        except NewslashDBError as e:
            db.rollback()
            raise e

        db.commit()
        db.close()
        return


    
