# -*- coding: utf-8 -*-
'''test for lucene_wrapper'''

import unittest
import calendar
from datetime import datetime
import shutil

from newslash_index import htmlutil
import lucene_wrapper

from test_config import config

item = {
    "time": datetime(2017, 10, 10, 10, 10, 10),
    "last_update": datetime(2017, 10, 10, 10, 10, 10),
    "introtext": u'''<p>イントロテキスト：<a href="http://example.com">example.com</a></p>''',
    "bodytext": "",
    "sid": "17/10/10/1234567",
    "title": u"テスト記事のタイトル",
    "dept": u"テストテストテスト",
    "tid": 2271,
    "uid": 27448,
    "submitter": 27448
}

class TestIndexClass(unittest.TestCase):
    def setUp(self):
        # cleanup existing index
        try:
            shutil.rmtree(config["SearchIndex"]["path"])
        except OSError:
            pass

    def tearDown(self):
        pass

    def test_index_manipulation(self):
        # convert datetime to UNIX timestamp
        timestamp = calendar.timegm(item["time"].utctimetuple())
        last_update = calendar.timegm(item["last_update"].utctimetuple())

        # prepare intro-/body-text, url
        introtext = item["introtext"] or ""
        bodytext = item["bodytext"] or ""
        (content_text, urls) = htmlutil.strip_html_tag(introtext + bodytext)

        # create index
        lucene_wrapper.init_vm()
        doc = lucene_wrapper.Document()

        doc.add_string_field("type", "story")
        doc.add_string_field("id", item["sid"])

        doc.add_text_field("title", item["title"])
        doc.add_text_field("content_text", content_text)
        doc.add_text_field("dept", item["dept"])

        doc.add_int_field("create_time", timestamp)
        doc.add_int_field("last_update", last_update)
        doc.add_int_field("topic", item["tid"])
        doc.add_int_field("author", item["uid"])
        doc.add_int_field("submitter", item["submitter"])

        for url in urls:
            doc.add_string_field("url", url)

        with lucene_wrapper.Indexer(index_directory=config["SearchIndex"]["path"]) as indexer:
            indexer.add(doc)

        # search
        item_type = "story"
        item_id = item["sid"]

        searcher = lucene_wrapper.Searcher(index_directory=config["SearchIndex"]["path"])
        type_query = lucene_wrapper.TermQuery("type",item_type)
        id_query = lucene_wrapper.TermQuery("id", item_id)

        query = lucene_wrapper.BooleanQuery()
        query.add_must(id_query)
        query.add_must(type_query)

        result = searcher.search(query)
        self.assertIs(result.total_hits, 1)

        # delete
        with lucene_wrapper.Indexer(index_directory=config["SearchIndex"]["path"]) as indexer:
            indexer.delete(query)

        searcher = lucene_wrapper.Searcher(index_directory=config["SearchIndex"]["path"])
        result = searcher.search(query)
        self.assertIs(result.total_hits, 0)
        
        # add new one
        new_text = u"更新されたテキスト"
        doc.remove_field("content_text")
        doc.add_text_field("content_text", new_text)
        
        with lucene_wrapper.Indexer(index_directory=config["SearchIndex"]["path"]) as indexer:
            indexer.add(doc)

        searcher = lucene_wrapper.Searcher(index_directory=config["SearchIndex"]["path"])
        result = searcher.search(query)
        self.assertIs(result.total_hits, 1)
        self.assertEqual(result[0].content_text, new_text)

        
if __name__ == '__main__':
    unittest.main()

