# -*- coding: utf-8 -*-
'''test for newslash_index'''

import unittest
from datetime import datetime
import shutil

from newslash_index import init_vm, Index, exceptions
import newslash_db
from test_config import config

TABLES_TO_USE = ("stories",
                 "story_text",
                 "story_param",
                 "ns_search_metadata",
                 "comments",
                 "comment_text",
                 "journals",
                 "journals_text",
                 "submissions",
                 "pollanswers",
                 "pollquestions",
                 "users",
                 "users_info",
                 "firehose",
)

TEST_MODE = 'all'

class TestIndexClass(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        # create metadata
        index = Index(config=config)
        try:
            index.metadata.create_table()
        except newslash_db.NewslashDBProgrammingError as e:
            pass

        # create tables
        for m in (newslash_db.Stories(config["Database"]),
                  newslash_db.Comments(config["Database"]),
                  newslash_db.Firehose(config["Database"]),
                  newslash_db.Journals(config["Database"]),
                  newslash_db.Submissions(config["Database"]),
                  newslash_db.Polls(config["Database"]),
                  newslash_db.Users(config["Database"]),
        ):
            m.create_tables()

        # check tables exist
        db = newslash_db.NewslashDB(config["Database"])
        #for name in TABLES_TO_USE:
        #    cur = db.execute('SHOW TABLES LIKE "{}"'.format(name))

        # cleanup existing index
        try:
            shutil.rmtree(config["SearchIndex"]["path"])
        except OSError:
            pass

        # init vm
        init_vm()

    @classmethod
    def tearDownClass(cls):
        # cleanup metadata table
        index = Index(config=config)
        index.metadata.drop_table()

        # cleanup tables
        for m in (newslash_db.Stories(config["Database"]),
                  newslash_db.Comments(config["Database"]),
                  newslash_db.Firehose(config["Database"]),
                  newslash_db.Journals(config["Database"]),
                  newslash_db.Submissions(config["Database"]),
                  newslash_db.Polls(config["Database"]),
                  newslash_db.Users(config["Database"]),
        ):
            try:
                m.drop_tables()
            except newslash_db.NewslashDBProgrammingError as e:
                pass

        # check tables not exist
        db = newslash_db.NewslashDB(config["Database"])
        #for name in TABLES_TO_USE:
        #    cur = db.execute('SHOW TABLES LIKE "{}"'.format(name))

        # cleanup existing index
        try:
            shutil.rmtree(config["SearchIndex"]["path"])
        except OSError:
            pass

    def setUp(self):
        self.index = Index(config=config)

    def tearDown(self):
        self.index.delete_all();

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_10create_metadata(self):
        with self.assertRaises(newslash_db.NewslashDBProgrammingError):
            self.index.metadata.create_table()

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_20update_comment_index(self):
        testdata = {
            "sid": 0,
            "pid": 0,
            "uid": 27448,
            "comment": u"コメントテスト"
        }

        # first, create test comment
        comments = newslash_db.Comments(config["Database"])
        cid = comments.insert(**testdata)
        self.assertIsNot(cid, 0)

        item = comments.select(cid=cid)
        self.assertIsNotNone(item)
        self.assertEqual(item[0].get("comment"), testdata["comment"])

        # create index
        suc, err = self.index.update_comment()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"コメント")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].type, "comment")
        self.assertEqual(docs[0].id, str(cid))

        # delete from index
        self.index.delete("comment", cid)
        
        # check index
        docs = self.index.query(u"コメント")
        self.assertIs(docs.total_hits, 0)
            

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_30update_story_index(self):
        testdata1 = {
            "sid": "17/10/10/1234567",
            "uid": 27448,
            "dept": u"テストテストテスト",
            "time": datetime(2017, 10, 10, 10, 10, 10),
            "title": u"テスト記事のタイトル",
            "introtext": u"<p>イントロテキスト：＜キーワード：猫＞</p>",
        }

        testdata2 = {
            "sid": "17/10/11/2345678",
            "uid": 27448,
            "dept": u"テストテストテスト2",
            "time": datetime(2017, 10, 11, 11, 11, 11),
            "title": u"2つ目のテスト記事のタイトル",
            "introtext": u"<p>イントロテキスト：＜キーワード：犬＞</p>",
        }

        # first, create test story
        stories = newslash_db.Stories(config["Database"])
        stoid1 = stories.insert(**testdata1)
        self.assertIsNot(stoid1, 0)

        item = stories.select(stoid=stoid1)
        self.assertIsNotNone(item)
        self.assertIsNot(item[0].get("neverdisplay"), "1")

        # create index
        suc, err = self.index.update_story()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, testdata1["sid"])

        # add new story
        stoid2 = stories.insert(**testdata2)
        self.assertIsNot(stoid2, 0)
                          
        # update index
        suc, err = self.index.update_story()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('story').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, testdata2["sid"])
        docs = self.index.query(u"テスト")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 2)

        # change story
        new_text = u"<p>猿</p>"
        testdata1["introtext"] = new_text
        testdata1["stoid"] = stoid1
        stories.update(**testdata1)
        item = stories.select(stoid=stoid1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["introtext"], new_text)

        # check updated stories
        items = stories.select(stoid_le=stoid2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_story()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('story').get('last_update')
        self.assertIsNotNone(last_update)
        
        # check updated index
        docs = self.index.query(u"猿")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, testdata1["sid"])
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # change story to neverdisplay
        testdata1["neverdisplay"] = "1"
        stories.update(**testdata1)
        item = stories.select(stoid=stoid1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["neverdisplay"], "1")

        # check updated stories
        items = stories.select(stoid_le=stoid2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_story()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('story').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"猿")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # delete all from index
        self.index.delete_all();
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 0)

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_40update_journal_index(self):
        testdata1 = {
            "uid": 27448,
            "description": u"テスト日記",
            "time": datetime(2017, 10, 10, 10, 10, 10),
            "article": u"<p>イントロテキスト：＜キーワード：猫＞</p>",
            "introtext": u"<p>イントロテキスト：＜キーワード：猫＞</p>",
            "public": "yes",
        }

        testdata2 = {
            "uid": 27448,
            "description": u"テスト日記",
            "time": datetime(2017, 11, 11, 11, 11, 11),
            "article": u"<p>イントロテキスト：＜キーワード：犬＞</p>",
            "introtext": u"<p>イントロテキスト：＜キーワード：犬＞</p>",
            "public": "yes",
        }

        # first, create test story
        journals = newslash_db.Journals(config["Database"])
        id1 = journals.insert(**testdata1)
        self.assertIsNot(id1, 0)
        testdata1["id"] = id1

        item = journals.select(id=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0].get("public"), "yes")

        # create index
        suc, err = self.index.update_journal()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata1["id"]))

        # add new story
        id2 = journals.insert(**testdata2)
        self.assertIsNot(id2, 0)
        testdata2["id"] = id2
                          
        # update index
        suc, err = self.index.update_journal()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('journal').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata2["id"]))
        docs = self.index.query(u"テスト")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 2)

        # change story
        new_text = u"<p>猿</p>"
        testdata1["article"] = new_text
        journals.update(**testdata1)
        item = journals.select(id=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["article"], new_text)

        # check updated journals
        items = journals.select(id_le=id2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_journal()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('journal').get('last_update')
        self.assertIsNotNone(last_update)
        
        # check updated index
        docs = self.index.query(u"猿")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata1["id"]))
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # change story to neverdisplay
        testdata1["public"] = "no"
        journals.update(**testdata1)
        item = journals.select(id=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["public"], "no")

        # check updated journals
        items = journals.select(id_le=id2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_journal()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('journal').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"猿")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # delete all from index
        self.index.delete_all();
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 0)

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_50update_submission_index(self):
        testdata1 = {
            "uid": 27448,
            "subj": u"テストタレコミ",
            "time": datetime(2017, 10, 10, 10, 10, 10),
            "story": u"<p>イントロテキスト：＜キーワード：猫＞</p>",
            "public": "yes",
            "tid": 1,
            "signature": "foobar",
        }

        testdata2 = {
            "uid": 27448,
            "subj": u"テストタレコミ",
            "time": datetime(2017, 11, 11, 11, 11, 11),
            "story": u"<p>イントロテキスト：＜キーワード：犬＞</p>",
            "public": "yes",
            "tid": 1,
            "signature": "hogehoge",
        }

        # first, create test story
        submissions = newslash_db.Submissions(config["Database"])
        id1 = submissions.insert(**testdata1)
        self.assertIsNot(id1, 0)
        testdata1["subid"] = id1

        item = submissions.select(subid=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0].get("public"), "yes")

        # create index
        suc, err = self.index.update_submission()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata1["subid"]))

        # add new story
        id2 = submissions.insert(**testdata2)
        self.assertIsNot(id2, 0)
        testdata2["subid"] = id2
                          
        # update index
        suc, err = self.index.update_submission()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('submission').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata2["subid"]))
        docs = self.index.query(u"テスト")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 2)

        # change story to neverdisplay
        testdata1["public"] = "no"
        submissions.update(**testdata1)
        item = submissions.select(subid=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["public"], "no")

        # check updated submissions
        items = submissions.select(subid_le=id2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_submission()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('submission').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # delete all from index
        self.index.delete_all();
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 0)

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_60update_poll_index(self):
        testdata1 = {
            "uid": 27448,
            "question": u"テスト投票",
            "date": datetime(2017, 10, 10, 10, 10, 10),
            "answers": [u"＜キーワード：猫＞", u"ほげほげ"],
            "topic": 1,
        }

        testdata2 = {
            "uid": 27448,
            "question": u"テスト投票",
            "date": datetime(2017, 11, 11, 11, 11, 11),
            "answers": [u"＜キーワード：犬＞", u"ほげほげ"],
            "topic": 1,
        }

        # first, create test story
        polls = newslash_db.Polls(config["Database"])
        id1 = polls.insert(**testdata1)
        self.assertIsNot(id1, 0)
        testdata1["qid"] = id1

        item = polls.select(qid=id1)
        self.assertIsNotNone(item)
        self.assertNotEqual(item[0].get("polltype"), "nodisplay")

        # create index
        suc, err = self.index.update_poll()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata1["qid"]))

        # add new story
        id2 = polls.insert(**testdata2)
        self.assertIsNot(id2, 0)
        testdata2["qid"] = id2
                          
        # update index
        suc, err = self.index.update_poll()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('poll').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata2["qid"]))
        docs = self.index.query(u"テスト")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 2)

        # change story to neverdisplay
        testdata1["polltype"] = "nodisplay"
        testdata1["date"] = datetime.now()
        polls.update(**testdata1)
        item = polls.select(qid=id1)
        self.assertIsNotNone(item)
        self.assertEqual(item[0]["polltype"], "nodisplay")

        # check updated polls
        items = polls.select(qid_le=id2, last_update_ge=last_update)
        self.assertTrue(len(items) > 0)

        # update index
        suc, err = self.index.update_poll()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('poll').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 1)
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)

        # delete all from index
        self.index.delete_all();
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 0)

    @unittest.skipIf(TEST_MODE != 'all', "test-mode is not all")
    def test_70update_user_index(self):
        testdata1 = {
            "nickname": u"テストユーザー1",
            "bio": u"＜猫＞",
            "homepage": "",
            "realname": "",
            "sig": u"キーワード",
        }

        testdata2 = {
            "nickname": u"テストユーザー2",
            "bio": u"＜犬＞",
            "homepage": "",
            "realname": "",
            "sig": u"キーワード",
        }

        # first, create test user
        users = newslash_db.Users(config["Database"])
        id1 = users.insert(**testdata1)
        self.assertIsNot(id1, 0)
        testdata1["uid"] = id1

        item = users.select(uid=id1)
        self.assertIsNotNone(item)

        # create index
        suc, err = self.index.update_user()
        self.assertIs(suc, 1)
        self.assertIs(err, 0)

        # check index
        docs = self.index.query(u"猫")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata1["uid"]))

        # add new user
        id2 = users.insert(**testdata2)
        self.assertIsNot(id2, 0)
        testdata2["uid"] = id2
                          
        # update index
        suc, err = self.index.update_user()
        self.assertIs(err, 0)
        self.assertIsNot(suc, 0)

        # check last_update
        last_update = self.index.metadata.get('user').get('last_update')
        self.assertIsNotNone(last_update)

        # check updated index
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 1)
        self.assertEqual(docs[0].id, str(testdata2["uid"]))
        docs = self.index.query(u"テスト")
        self.assertIs(docs.total_hits, 0)
        docs = self.index.query(u"キーワード")
        self.assertIs(docs.total_hits, 2)

        # delete all from index
        self.index.delete_all();
        docs = self.index.query(u"犬")
        self.assertIs(docs.total_hits, 0)

if __name__ == '__main__':
    unittest.main()
    
