#!/usr/bin/python
"""newslash's statics exporter for prometheus"""

try:
    from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer
    from urlparse import parse_qs, urlparse
except ImportError:
    # Python 3
    from http.server import BaseHTTPRequestHandler, HTTPServer
    from urllib.parse import parse_qs, urlparse

from datetime import datetime, timedelta
import argparse

import redis
from prometheus_client import core, Gauge, generate_latest

TARGET_CODES = [200,
                301, 302, 304, 307, 308,
                400, 401, 403, 404, 405, 406, 408, 409,
                500]
#TARGET_CODES.extend(range(100, 103))
#TARGET_CODES.extend(range(200, 209))
#TARGET_CODES.extend(range(300, 309))
#TARGET_CODES.extend(range(400, 424))
#TARGET_CODES.extend([426, 451])
#TARGET_CODES.extend(range(500, 511))

TARGET_METHODS = ["GET", "POST"]

PORT = 9350
ADDRESS = '';
REQUEST_PER_SECONDS = Gauge('newslash_reuqest_per_seconds', 'HTTP(S) Request per seconds', ['method', 'code'])
KEY_PREFIX = "newslash_statics:"
KEY_REQUEST_COUNTER = "req_count"
REQUEST_COUNTER_BASE_TIME_FIELD = "base_time"

ARGUMENTS = {}

CONTENT_TYPE_LATEST = str('text/plain; version=0.0.4; charset=utf-8')
'''Content type of the latest text format'''

class MetricsHandler(BaseHTTPRequestHandler):
    """HTTP handler that gives metrics from ``core.REGISTRY``."""

    def do_GET(self):
        self._build_metrics()
        registry = core.REGISTRY
        params = parse_qs(urlparse(self.path).query)
        if 'name[]' in params:
            registry = registry.restricted_registry(params['name[]'])
        try:
            output = generate_latest(registry)
        except:
            self.send_error(500, 'error generating metric output')
            raise
        self.send_response(200)
        self.send_header('Content-Type', CONTENT_TYPE_LATEST)
        self.end_headers()
        self.wfile.write(output)
            
    def log_message(self, format, *args):
        """Log nothing."""

    def _build_metrics(self):
        host = ARGUMENTS.redis_host
        port = ARGUMENTS.redis_port or 6379
        key = KEY_PREFIX + KEY_REQUEST_COUNTER

        r = redis.StrictRedis(host=host, port=port)
        statics = r.hgetall(key)
        time = r.time()

        try: 
            base_time = float(statics[REQUEST_COUNTER_BASE_TIME_FIELD])
            current_time = float(time[0])
        except KeyError:
            return
        except ValueError:
            p = r.pipeline()
            p.delete(key)
            p.hset(key, REQUEST_COUNTER_BASE_TIME_FIELD, time[0])
            p.execute()
            return

        base_dt = datetime.fromtimestamp(base_time)
        current_dt = datetime.fromtimestamp(current_time)
        t_delta = current_dt - base_dt
        duration_sec = float(t_delta.total_seconds())

        # reset all values
        REQUEST_PER_SECONDS.labels("other", "other").set(0)
        for a in TARGET_METHODS:
            REQUEST_PER_SECONDS.labels(a, "other").set(0)
            for c in TARGET_CODES:
                REQUEST_PER_SECONDS.labels(a, c).set(0)

        for f in statics:
            if f == REQUEST_COUNTER_BASE_TIME_FIELD:
                continue
            items = f.split(":")
            try:
                method = items[0]
                code = int(items[1])
                value = float(statics[f])
                value = value / duration_sec
            except (IndexError, ValueError):
                return

            if code not in TARGET_CODES:
                code = "other"
            if method not in TARGET_METHODS:
                method = "other"

            REQUEST_PER_SECONDS.labels(method, code).set(value)

        p = r.pipeline()
        p.delete(key)
        p.hset(key, REQUEST_COUNTER_BASE_TIME_FIELD, time[0])
        p.execute()

if __name__ == '__main__':
    parser = argparse.ArgumentParser(description='newslash\'s statics exporter for prometheus')
    parser.add_argument('--redis-host', dest='redis_host', help='redis host name', required=True)
    parser.add_argument('--redis-port', dest='redis_port', help='redis port number (default:6379)')
    args = parser.parse_args()
    ARGUMENTS = args

    httpd = HTTPServer((ADDRESS, PORT), MetricsHandler)
    print "server starting at" + ADDRESS + ":" + str(PORT)
    httpd.serve_forever()

    
