#!/usr/bin/perl
use strict;
use warnings;
use utf8;
use feature ':5.10';

package DBUtil;
use DBI;
use File::Basename;
use File::Spec;


=encoding utf8

=head1 NAME

DBUtil - Database utility class for newslash

=head1 SYNOPSIS

  use DBUtil;

=head1 METHODS

=cut



=head2 new

create new DBUtil instance.

  my $db = DBUtil->new(host => "foo",
                       user => "bar",
                       password => "hoge",
                       database => "moge",
                       table_sql_directory => "foofoo");
                       data_sql_directory => "foobar");

=cut

sub new {
    my ($class, @rest) = @_;
    my $options = {@rest};

    bless {options => $options,
           _error => "",
           _errorno => 0,
          }, $class;
}


=head2 options

=head2 host

=head2 user

=head2 password

=head2 database

=head2 sql_directory

=head2 data_sql_directory

get each properties (getter)

  $db->options;
  $db->host;
  $db->user;
  $db->password;
  $db->database;
  $db->table_sql_directory;
  $db->data_sql_directory;

=cut

sub options {shift->{options}}
sub host {shift->options->{host}}
sub user {shift->options->{user}}
sub password {shift->options->{password}}
sub database {shift->options->{database}}
sub table_sql_directory {shift->options->{table_sql_directory}}
sub data_sql_directory {shift->options->{data_sql_directory}}

=head2 initialize_database

Initialize database. Create tables, then insert some vital datas.

  $db->initialize_database;

=cut

sub initialize_database {
    my $self = shift;

    my $dbh = $self->_connect;
    $dbh->{AutoCommit} = 0;

    # create tables
    if (!$self->_load_and_run_sql($self->table_sql_directory, $dbh)) {
        $dbh->rollback();
        die "cannot create tables correctly";
    }

    # insert data
    if (!$self->_load_and_run_sql($self->data_sql_directory, $dbh)) {
        $dbh->rollback();
        die "cannot insert datas correctly";
    }

    $dbh->commit;
    $dbh->disconnect;
    return 1;
}

=head2 delete_tables

Delete all tables in database.

  $db->delete_tables

=cut

sub delete_tables {
    my $self = shift;

    my $dbh = $self->_connect;
    # get current tables
    my $sth = $dbh->prepare("SHOW TABLES");
    $sth->execute or die "cannot retrieve table data";

    my $rs = $sth->fetchall_arrayref([0]);
    for my $name (@$rs) {
        if (!$dbh->do("DROP TABLE `$name->[0]`")) {
            warn "cannot delete table $name->[0]\n";
        }
    }
}


sub _load_and_run_sql {
    my ($self, $dirname, $dbh) = @_;

    opendir(my $dh, $dirname) || die "cannot open SQL directory: $dirname";

    my @files;
    while (my $fname = readdir $dh) {
        if ($fname =~ m/^\./) {
            next;
        }
        if ($fname =~ m/^.*\.sql$/) {
            push @files, $fname;
        }
    }
    @files = sort @files;
    while (my $fname = shift @files) {
        my $sql = $self->_get_sql_contents($dirname, $fname);
        my $result = $dbh->do($sql);
        if (!$result) {
            warn "$dirname / $fname : sql execute error - $!\n";
            return;
        }
    }
    return 1;
}

sub _get_sql_contents {
    my ($self, $dirname, $fname) = @_;

    my $pathname = File::Spec->catfile($dirname, $fname);
    my $fh;

    if (!open($fh, '<', $pathname)) {
        print "cannot open $fh: $!\n";
        return;
    }

    my $sql = do { local $/; <$fh> };
    return $sql;
}

# connect database
sub _connect {
    my $self = shift;
    my $host = $self->host;
    my $database = $self->database;
    my $dbh = DBI->connect("DBI:mysql:$database:$host", $self->user, $self->password);
    if (!$dbh) { die "cannot connet database!\n" }
    return $dbh;
}


1;
